# sage_setup: distribution = sagemath-combinat
"""
Derangements

AUTHORS:

- Alasdair McAndrew (2010-05): Initial version
- Travis Scrimshaw (2013-03-30): Put derangements into category framework
"""

# ****************************************************************************
#       Copyright (C) 2010 Alasdair McAndrew <amca01@gmail.com>,
#                     2013 Travis Scrimshaw <tscrim@ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.structure.parent import Parent
from sage.structure.unique_representation import UniqueRepresentation
from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets
from sage.misc.misc_c import prod
from sage.misc.prandom import random, randrange
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ
from sage.rings.integer import Integer
from sage.combinat.combinat import CombinatorialElement
from sage.combinat.permutation import Permutation, Permutations


class Derangement(CombinatorialElement):
    r"""
    A derangement.

    A derangement on a set `S` is a permutation `\sigma` such that `\sigma(x)
    \neq x` for all `x \in S`, i.e. `\sigma` is a permutation of `S` with no
    fixed points.

    EXAMPLES::

        sage: D = Derangements(4)
        sage: elt = D([4,3,2,1])
        sage: TestSuite(elt).run()
    """

    def to_permutation(self):
        """
        Return the permutation corresponding to ``self``.

        EXAMPLES::

            sage: D = Derangements(4)
            sage: p = D([4,3,2,1]).to_permutation(); p
            [4, 3, 2, 1]
            sage: type(p)
            <class 'sage.combinat.permutation.StandardPermutations_all_with_category.element_class'>
            sage: D = Derangements([1, 3, 3, 4])
            sage: D[0].to_permutation()
            Traceback (most recent call last):
            ...
            ValueError: can only convert to a permutation for derangements of [1, 2, ..., n]
        """
        if self.parent()._set != tuple(range(1, len(self) + 1)):
            raise ValueError("can only convert to a permutation for derangements of [1, 2, ..., n]")
        return Permutation(list(self))


class Derangements(UniqueRepresentation, Parent):
    r"""
    The class of all derangements of a set or multiset.

    A derangement on a set `S` is a permutation `\sigma` such that `\sigma(x)
    \neq x` for all `x \in S`, i.e. `\sigma` is a permutation of `S` with no
    fixed points.

    For an integer, or a list or string with all elements
    distinct, the derangements are obtained by a standard result described
    in [BV2004]_. For a list or string with repeated elements, the derangements
    are formed by computing all permutations of the input and discarding all
    non-derangements.

    INPUT:

    - ``x`` -- can be an integer which corresponds to derangements of
      `\{1, 2, 3, \ldots, x\}`, a list, or a string

    REFERENCES:

    - [BV2004]_
    - :wikipedia:`Derangement`

    EXAMPLES::

        sage: D1 = Derangements([2,3,4,5])
        sage: D1.list()
        [[3, 4, 5, 2],
         [5, 4, 2, 3],
         [3, 5, 2, 4],
         [4, 5, 3, 2],
         [4, 2, 5, 3],
         [5, 2, 3, 4],
         [5, 4, 3, 2],
         [4, 5, 2, 3],
         [3, 2, 5, 4]]
        sage: D1.cardinality()
        9
        sage: D1.random_element() # random
        [4, 2, 5, 3]
        sage: D2 = Derangements([1,2,3,1,2,3])
        sage: D2.cardinality()
        10
        sage: D2.list()
        [[2, 1, 1, 3, 3, 2],
         [2, 1, 2, 3, 3, 1],
         [2, 3, 1, 2, 3, 1],
         [2, 3, 1, 3, 1, 2],
         [2, 3, 2, 3, 1, 1],
         [3, 1, 1, 2, 3, 2],
         [3, 1, 2, 2, 3, 1],
         [3, 1, 2, 3, 1, 2],
         [3, 3, 1, 2, 1, 2],
         [3, 3, 2, 2, 1, 1]]
        sage: D2.random_element() # random
        [2, 3, 1, 3, 1, 2]
    """
    @staticmethod
    def __classcall_private__(cls, x):
        """
        Normalize ``x`` to ensure a unique representation.

        EXAMPLES::

            sage: D = Derangements(4)
            sage: D2 = Derangements([1, 2, 3, 4])
            sage: D3 = Derangements((1, 2, 3, 4))
            sage: D is D2
            True
            sage: D is D3
            True
        """
        if x in ZZ:
            x = tuple(range(1, x + 1))
        return super().__classcall__(cls, tuple(x))

    def __init__(self, x):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: D = Derangements(4)
            sage: TestSuite(D).run()
            sage: D = Derangements('abcd')
            sage: TestSuite(D).run()
            sage: D = Derangements([2, 2, 1, 1])
            sage: TestSuite(D).run()
        """
        Parent.__init__(self, category=FiniteEnumeratedSets())
        self._set = x
        self.__multi = len(set(x)) < len(x)

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: Derangements(4)
            Derangements of the set [1, 2, 3, 4]
            sage: Derangements('abcd')
            Derangements of the set ['a', 'b', 'c', 'd']
            sage: Derangements([2,2,1,1])
            Derangements of the multiset [2, 2, 1, 1]
        """
        if self.__multi:
            return "Derangements of the multiset %s" % list(self._set)
        return "Derangements of the set %s" % list(self._set)

    def _element_constructor_(self, der):
        """
        Construct an element of ``self`` from ``der``.

        EXAMPLES::

            sage: D = Derangements(4)
            sage: elt = D([3,1,4,2]); elt
            [3, 1, 4, 2]
            sage: elt.parent() is D
            True
        """
        if isinstance(der, Derangement):
            if der.parent() is self:
                return der
            raise ValueError("cannot convert %s to an element of %s" % (der, self))
        return self.element_class(self, der)

    Element = Derangement

    def __iter__(self):
        """
        Iterate through ``self``.

        EXAMPLES::

            sage: D = Derangements(4)
            sage: D.list() # indirect doctest
            [[2, 3, 4, 1],
             [4, 3, 1, 2],
             [2, 4, 1, 3],
             [3, 4, 2, 1],
             [3, 1, 4, 2],
             [4, 1, 2, 3],
             [4, 3, 2, 1],
             [3, 4, 1, 2],
             [2, 1, 4, 3]]
            sage: D = Derangements([1,44,918,67])
            sage: D.list()
            [[44, 918, 67, 1],
             [67, 918, 1, 44],
             [44, 67, 1, 918],
             [918, 67, 44, 1],
             [918, 1, 67, 44],
             [67, 1, 44, 918],
             [67, 918, 44, 1],
             [918, 67, 1, 44],
             [44, 1, 67, 918]]
            sage: D = Derangements(['A','AT','CAT','CATS'])
            sage: D.list()
            [['AT', 'CAT', 'CATS', 'A'],
             ['CATS', 'CAT', 'A', 'AT'],
             ['AT', 'CATS', 'A', 'CAT'],
             ['CAT', 'CATS', 'AT', 'A'],
             ['CAT', 'A', 'CATS', 'AT'],
             ['CATS', 'A', 'AT', 'CAT'],
             ['CATS', 'CAT', 'AT', 'A'],
             ['CAT', 'CATS', 'A', 'AT'],
             ['AT', 'A', 'CATS', 'CAT']]
            sage: D = Derangements('CART')
            sage: D.list()
            [['A', 'R', 'T', 'C'],
             ['T', 'R', 'C', 'A'],
             ['A', 'T', 'C', 'R'],
             ['R', 'T', 'A', 'C'],
             ['R', 'C', 'T', 'A'],
             ['T', 'C', 'A', 'R'],
             ['T', 'R', 'A', 'C'],
             ['R', 'T', 'C', 'A'],
             ['A', 'C', 'T', 'R']]
            sage: D = Derangements([1,1,2,2,3,3])
            sage: D.list()
            [[2, 2, 3, 3, 1, 1],
             [2, 3, 1, 3, 1, 2],
             [2, 3, 1, 3, 2, 1],
             [2, 3, 3, 1, 1, 2],
             [2, 3, 3, 1, 2, 1],
             [3, 2, 1, 3, 1, 2],
             [3, 2, 1, 3, 2, 1],
             [3, 2, 3, 1, 1, 2],
             [3, 2, 3, 1, 2, 1],
             [3, 3, 1, 1, 2, 2]]
            sage: D = Derangements('SATTAS')
            sage: D.list()
            [['A', 'S', 'S', 'A', 'T', 'T'],
             ['A', 'S', 'A', 'S', 'T', 'T'],
             ['A', 'T', 'S', 'S', 'T', 'A'],
             ['A', 'T', 'S', 'A', 'S', 'T'],
             ['A', 'T', 'A', 'S', 'S', 'T'],
             ['T', 'S', 'S', 'A', 'T', 'A'],
             ['T', 'S', 'A', 'S', 'T', 'A'],
             ['T', 'S', 'A', 'A', 'S', 'T'],
             ['T', 'T', 'S', 'A', 'S', 'A'],
             ['T', 'T', 'A', 'S', 'S', 'A']]
            sage: D = Derangements([1,1,2,2,2])
            sage: D.list()
            []
            sage: D = Derangements(0)
            sage: D.list()
            [[]]
        """
        if self.__multi:
            for p in Permutations(self._set):
                if not self._fixed_point(p):
                    yield self.element_class(self, list(p))
        else:
            for d in self._iter_der(len(self._set)):
                yield self.element_class(self, [self._set[i - 1] for i in d])

    def _iter_der(self, n):
        r"""
        Iterate through all derangements of the list `[1, 2, 3, \ldots, n]`
        using the method given in [BV2004]_.

        EXAMPLES::

            sage: D = Derangements(4)
            sage: list(D._iter_der(4))
            [[2, 3, 4, 1],
             [4, 3, 1, 2],
             [2, 4, 1, 3],
             [3, 4, 2, 1],
             [3, 1, 4, 2],
             [4, 1, 2, 3],
             [4, 3, 2, 1],
             [3, 4, 1, 2],
             [2, 1, 4, 3]]
        """
        if n == 0:
            yield []
            return
        elif n == 1:
            return
        elif n == 2:
            yield [2, 1]
        elif n == 3:
            yield [2, 3, 1]
            yield [3, 1, 2]
        elif n >= 4:
            for d in self._iter_der(n - 1):
                for i in range(1, n):
                    s = d[:]
                    ii = d.index(i)
                    s[ii] = n
                    yield s + [i]
            for d in self._iter_der(n - 2):
                for i in range(1, n):
                    s = d[:]
                    s = [x >= i and x + 1 or x for x in s]
                    s.insert(i - 1, n)
                    yield s + [i]

    def _fixed_point(self, a):
        """
        Return ``True`` if ``a`` has a point in common with ``self._set``.

        EXAMPLES::

            sage: D = Derangements(5)
            sage: D._fixed_point([3,1,2,5,4])
            False
            sage: D._fixed_point([5,4,3,2,1])
            True
        """
        return any(x == y for (x, y) in zip(a, self._set))

    def _count_der(self, n):
        """
        Count the number of derangements of `n` using the recursion
        `D_2 = 1, D_3 = 2, D_n = (n-1) (D_{n-1} + D_{n-2})`.

        EXAMPLES::

            sage: D = Derangements(5)
            sage: D._count_der(2)
            1
            sage: D._count_der(3)
            2
            sage: D._count_der(5)
            44
        """
        if n == 0:
            return Integer(1)
        if n == 1:
            return Integer(0)
        if n == 2:
            return Integer(1)
        if n == 3:
            return Integer(2)
        # n >= 4
        last = Integer(2)
        second_last = Integer(1)
        for i in range(4, n + 1):
            current = (i - 1) * (last + second_last)
            second_last = last
            last = current
        return last

    def cardinality(self):
        r"""
        Counts the number of derangements of a positive integer, a
        list, or a string.  The list or string may contain repeated
        elements.  If an integer `n` is given, the value returned
        is the number of derangements of `[1, 2, 3, \ldots, n]`.

        For an integer, or a list or string with all elements
        distinct, the value is obtained by the standard result
        `D_2 = 1, D_3 = 2, D_n = (n-1) (D_{n-1} + D_{n-2})`.

        For a list or string with repeated elements, the number of
        derangements is computed by Macmahon's theorem. If the numbers
        of repeated elements are `a_1, a_2, \ldots, a_k` then the number
        of derangements is given by the coefficient of `x_1 x_2 \cdots
        x_k` in the expansion of `\prod_{i=0}^k (S - s_i)^{a_i}` where
        `S = x_1 + x_2 + \cdots + x_k`.

        EXAMPLES::

            sage: D = Derangements(5)
            sage: D.cardinality()
            44
            sage: D = Derangements([1,44,918,67,254])
            sage: D.cardinality()
            44
            sage: D = Derangements(['A','AT','CAT','CATS','CARTS'])
            sage: D.cardinality()
            44
            sage: D = Derangements('UNCOPYRIGHTABLE')
            sage: D.cardinality()
            481066515734
            sage: D = Derangements([1,1,2,2,3,3])
            sage: D.cardinality()
            10
            sage: D = Derangements('SATTAS')
            sage: D.cardinality()
            10
            sage: D = Derangements([1,1,2,2,2])
            sage: D.cardinality()
            0
            sage: D = Derangements(0)
            sage: D.cardinality()
            1
        """
        if self.__multi:
            sL = set(self._set)
            A = [self._set.count(i) for i in sL]
            R = PolynomialRing(QQ, 'x', len(A))
            S = sum(R.gens())
            e = prod((S - x)**y for (x, y) in zip(R.gens(), A))
            return Integer(e.coefficient(dict(zip(R.gens(), A))))
        return self._count_der(len(self._set))

    def _rand_der(self):
        r"""
        Produces a random derangement of `[1, 2, \ldots, n]`.

        This is an
        implementation of the algorithm described by Martinez et. al. in
        [MPP2008]_.

        EXAMPLES::

            sage: D = Derangements(4)
            sage: d = D._rand_der()
            sage: d in D
            True
        """
        n = len(self._set)
        A = list(range(1, n + 1))
        mark = [x < 0 for x in A]
        i, u = n, n
        while u >= 2:
            if not mark[i - 1]:
                while True:
                    j = randrange(1, i)
                    if not mark[j - 1]:
                        A[i - 1], A[j - 1] = A[j - 1], A[i - 1]
                        break
                p = random()
                if p < (u - 1) * self._count_der(u - 2) // self._count_der(u):
                    mark[j - 1] = True
                    u -= 1
                u -= 1
            i -= 1
        return A

    def random_element(self):
        r"""
        Produce all derangements of a positive integer, a list, or
        a string.  The list or string may contain repeated elements.
        If an integer `n` is given, then a random
        derangements of `[1, 2, 3, \ldots, n]` is returned

        For an integer, or a list or string with all elements
        distinct, the value is obtained by an algorithm described in
        [MPP2008]_. For a list or string with repeated elements the
        derangement is formed by choosing an element at random from the list of
        all possible derangements.

        OUTPUT:

        A single list or string containing a derangement, or an
        empty list if there are no derangements.

        EXAMPLES::

            sage: D = Derangements(4)
            sage: D.random_element() # random
            [2, 3, 4, 1]
            sage: D = Derangements(['A','AT','CAT','CATS','CARTS','CARETS'])
            sage: D.random_element() # random
            ['AT', 'CARTS', 'A', 'CAT', 'CARETS', 'CATS']
            sage: D = Derangements('UNCOPYRIGHTABLE')
            sage: D.random_element() # random
            ['C', 'U', 'I', 'H', 'O', 'G', 'N', 'B', 'E', 'L', 'A', 'R', 'P', 'Y', 'T']
            sage: D = Derangements([1,1,1,1,2,2,2,2,3,3,3,3])
            sage: D.random_element() # random
            [3, 2, 2, 3, 1, 3, 1, 3, 2, 1, 1, 2]
            sage: D = Derangements('ESSENCES')
            sage: D.random_element() # random
            ['N', 'E', 'E', 'C', 'S', 'S', 'S', 'E']
            sage: D = Derangements([1,1,2,2,2])
            sage: D.random_element()
            []

        TESTS:

        Check that index error discovered in :issue:`29974` is fixed::

            sage: D = Derangements([1,1,2,2])
            sage: _ = [D.random_element() for _ in range(20)]
        """
        if self.__multi:
            L = list(self)
            if len(L) == 0:
                return self.element_class(self, [])
            i = randrange(len(L))
            return L[i]
        temp = self._rand_der()
        return self.element_class(self, [self._set[ii - 1] for ii in temp])



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.groups sage.modules
r"""
Specht modules

.. TODO::

    Integrate this with the implementations in
    :mod:`sage.modules.with_basis.representation`.

AUTHORS:

- Travis Scrimshaw (2023-1-22): initial version
- Travis Scrimshaw (2023-11-23): added simple modules based on code
  from Sacha Goldman
"""

# ****************************************************************************
#       Copyright (C) 2023 Travis Scrimshaw <tcscrims (at) gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.cachefunc import cached_method
from sage.misc.lazy_attribute import lazy_attribute
from sage.combinat.diagram import Diagram
from sage.combinat.partition import _Partitions
from sage.combinat.free_module import CombinatorialFreeModule
from sage.modules.with_basis.representation import Representation_abstract
from sage.sets.family import Family
from sage.matrix.constructor import matrix
from sage.rings.rational_field import QQ
from sage.modules.with_basis.subquotient import SubmoduleWithBasis, QuotientModuleWithBasis
from sage.categories.modules_with_basis import ModulesWithBasis


class SymmetricGroupRepresentation(Representation_abstract):
    """
    Mixin class for symmetric group (algebra) representations.
    """
    def __init__(self, SGA):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: SM = Partition([3,1,1]).specht_module(GF(3))
            sage: SM.side()
            'left'
            sage: TestSuite(SM).run()
        """
        Representation_abstract.__init__(self, SGA.group(), "left", SGA)

    @cached_method
    def frobenius_image(self):
        r"""
        Return the Frobenius image of ``self``.

        The Frobenius map is defined as the map to symmetric functions

        .. MATH::

            F(\chi) = \frac{1}{n!} \sum_{w \in S_n} \chi(w) p_{\rho(w)},

        where `\chi` is the character of the `S_n`-module ``self``,
        `p_{\lambda}` is the powersum symmetric function basis element
        indexed by `\lambda`, and `\rho(w)` is the cycle type of `w` as a
        partition. Specifically, this map takes irreducible representations
        indexed by `\lambda` to the Schur function `s_{\lambda}`.

        EXAMPLES::

            sage: SM = Partition([2,2,1]).specht_module(QQ)
            sage: SM.frobenius_image()
            s[2, 2, 1]
            sage: SM = Partition([4,1]).specht_module(CyclotomicField(5))
            sage: SM.frobenius_image()
            s[4, 1]

        We verify the regular representation::

            sage: from sage.combinat.diagram import Diagram
            sage: D = Diagram([(0,0), (1,1), (2,2), (3,3), (4,4)])
            sage: F = D.specht_module(QQ).frobenius_image(); F
            s[1, 1, 1, 1, 1] + 4*s[2, 1, 1, 1] + 5*s[2, 2, 1]
             + 6*s[3, 1, 1] + 5*s[3, 2] + 4*s[4, 1] + s[5]
            sage: s = SymmetricFunctions(QQ).s()
            sage: F == sum(StandardTableaux(la).cardinality() * s[la]
            ....:          for la in Partitions(5))
            True
            sage: all(s[la] == la.specht_module(QQ).frobenius_image()
            ....:     for n in range(1, 5) for la in Partitions(n))
            True

            sage: D = Diagram([(0,0), (1,1), (1,2), (2,3), (2,4)])
            sage: SM = D.specht_module(QQ)
            sage: SM.frobenius_image()
            s[2, 2, 1] + s[3, 1, 1] + 2*s[3, 2] + 2*s[4, 1] + s[5]

        An example using the tabloid module::

            sage: SGA = SymmetricGroupAlgebra(QQ, 5)
            sage: TM = SGA.tabloid_module([2, 2, 1])
            sage: TM.frobenius_image()
            s[2, 2, 1] + s[3, 1, 1] + 2*s[3, 2] + 2*s[4, 1] + s[5]
        """
        from sage.combinat.sf.sf import SymmetricFunctions
        p = SymmetricFunctions(QQ).p()
        s = SymmetricFunctions(QQ).s()
        G = self._semigroup
        CCR = [(elt, elt.cycle_type()) for elt in G.conjugacy_classes_representatives()]
        B = self.basis()
        return s(p.sum(QQ(sum((elt * B[k])[k] for k in B.keys())) / la.centralizer_size() * p[la]
                       for elt, la in CCR))


class SpechtModule(SymmetricGroupRepresentation, SubmoduleWithBasis):
    r"""
    A Specht module.

    Let `S_n` be the symmetric group on `n` letters and `R` be a commutative
    ring. The *Specht module* `S^D` for a diagram `D` is an `S_n`-module
    defined as follows. Let

    .. MATH::

        R(D) := \sum_{w \in R_D} w,
        \qquad\qquad
        C(D) := \sum_{w \in C_D} (-1)^w w,

    where `R_D` (resp. `C_D`) is the row (resp. column) stabilizer of `D`.
    Then, we construct the Specht module `S^D` as the left ideal

    .. MATH::

        S^D = R[S_n] C(D) R(D),

    where `R[S_n]` is the group algebra of `S_n` over `R`.

    INPUT:

    - ``SGA`` -- a symmetric group algebra
    - ``D`` -- a diagram

    EXAMPLES:

    We begin by constructing all irreducible Specht modules for the symmetric
    group `S_4` and show that they give a full set of irreducible
    representations both by having distinct Frobenius characters and the
    sum of the square of their dimensions is equal to `4!`::

        sage: SP = [la.specht_module(QQ) for la in Partitions(4)]
        sage: s = SymmetricFunctions(QQ).s()
        sage: [s(S.frobenius_image()) for S in SP]
        [s[4], s[3, 1], s[2, 2], s[2, 1, 1], s[1, 1, 1, 1]]
        sage: sum(S.dimension()^2 for S in SP)
        24

    Next, we compute the Specht module for a more general diagram
    for `S_5` and compute its irreducible decomposition by using
    its Frobenius character::

        sage: D = [(0,0), (0,1), (1,1), (1,2), (0,3)]
        sage: SGA = SymmetricGroupAlgebra(QQ, 5)
        sage: SM = SGA.specht_module(D)
        sage: SM.dimension()
        9
        sage: s(SM.frobenius_image())
        s[3, 2] + s[4, 1]

    This carries a natural (left) action of the symmetric group (algebra)::

        sage: S5 = SGA.group()
        sage: v = SM.an_element(); v
        2*S[0] + 2*S[1] + 3*S[2]
        sage: S5([2,1,5,3,4]) * v
        3*S[0] + 2*S[1] + 2*S[2]
        sage: x = SGA.an_element(); x
        [1, 2, 3, 4, 5] + 2*[1, 2, 3, 5, 4] + 3*[1, 2, 4, 3, 5] + [5, 1, 2, 3, 4]
        sage: x * v
        15*S[0] + 14*S[1] + 16*S[2] - 7*S[5] + 2*S[6] + 2*S[7]

    .. SEEALSO::

        :class:`~sage.combinat.symmetric_group_representations.SpechtRepresentation`
        for an implementation of the representation by matrices.
    """
    @staticmethod
    def __classcall_private__(cls, SGA, D):
        r"""
        Normalize input to ensure a unique representation.

        EXAMPLES::

            sage: from sage.combinat.specht_module import SpechtModule
            sage: from sage.combinat.diagram import Diagram
            sage: SGA = SymmetricGroupAlgebra(QQ, 3)
            sage: D = [(0,0), (1,1), (1,2)]
            sage: SM1 = SpechtModule(SGA, D)
            sage: SM2 = SpechtModule(SGA, Diagram(D))
            sage: SM1 is SM2
            True
            sage: SM1 is SpechtModule(SGA, [[1,1], [1,2], [0,0]])
            True

            sage: SpechtModule(SGA, [[0,0], [1,1]])
            Traceback (most recent call last):
            ...
            ValueError: the domain size (=3) does not match the number of boxes (=2) of the diagram
        """
        if D in _Partitions:
            D = _Partitions(D)
            return TabloidModule(SGA, D).specht_module()
        D = _to_diagram(D)
        D = Diagram(D)
        n = len(D)
        if SGA.group().rank() != n - 1:
            rk = SGA.group().rank() + 1
            raise ValueError(f"the domain size (={rk}) does not match the number of boxes (={n}) of the diagram")
        return super().__classcall__(cls, SGA, D)

    def __init__(self, SGA, D):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(QQ, 4)
            sage: SM = SGA.specht_module([(0,0), (1,1), (1,2), (2,1)])
            sage: TestSuite(SM).run()
        """
        SymmetricGroupRepresentation.__init__(self, SGA)
        self._diagram = D
        Mod = ModulesWithBasis(SGA.category().base_ring())
        span_set = specht_module_spanning_set(D, SGA)
        support_order = SGA.get_order()
        basis = SGA.echelon_form(span_set, False, order=support_order)
        basis = Family(basis)
        SubmoduleWithBasis.__init__(self, basis, support_order, ambient=SGA,
                                    unitriangular=False, category=Mod.Subobjects(),
                                    prefix='S')

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(QQ, 4)
            sage: SGA.specht_module([(0,0), (1,1), (1,2), (2,1)])
            Specht module of [(0, 0), (1, 1), (1, 2), (2, 1)] over Rational Field

            sage: SGA.specht_module([3, 1])
            Specht module of [3, 1] over Rational Field
        """
        return f"Specht module of {self._diagram} over {self.base_ring()}"

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(QQ, 4)
            sage: SM = SGA.specht_module([(0,0), (1,1), (1,2), (2,1)])
            sage: latex(SM)
            S^{{\def\lr#1{\multicolumn{1}{|@{\hspace{.6ex}}c@{\hspace{.6ex}}|}{\raisebox{-.3ex}{$#1$}}}
            \raisebox{-.6ex}{$\begin{array}[b]{*{3}{p{0.6ex}}}\cline{1-1}
            \lr{\phantom{x}}&&\\\cline{1-1}\cline{2-2}\cline{3-3}
            &\lr{\phantom{x}}&\lr{\phantom{x}}\\\cline{2-2}\cline{3-3}\cline{2-2}
            &\lr{\phantom{x}}&\\\cline{2-2}
            \end{array}$}
            }}

            sage: SM = SGA.specht_module([3,1])
            sage: latex(SM)
            S^{{\def\lr#1{\multicolumn{1}{|@{\hspace{.6ex}}c@{\hspace{.6ex}}|}{\raisebox{-.3ex}{$#1$}}}
            \raisebox{-.6ex}{$\begin{array}[b]{*{3}c}\cline{1-3}
            \lr{\phantom{x}}&\lr{\phantom{x}}&\lr{\phantom{x}}\\\cline{1-3}
            \lr{\phantom{x}}\\\cline{1-1}
            \end{array}$}
            }}
        """
        from sage.misc.latex import latex
        return f"S^{{{latex(self._diagram)}}}"

    def _ascii_art_(self):
        r"""
        Return an ascii art representation of ``self``.

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(QQ, 4)
            sage: SM = SGA.specht_module([(0,0), (1,1), (1,2), (2,1)])
            sage: ascii_art(SM)
             O . .
             . O O
             . O .
            S

            sage: SM = SGA.specht_module([3,1])
            sage: ascii_art(SM)
             ***
             *
            S
        """
        from sage.typeset.ascii_art import ascii_art
        return ascii_art("S", baseline=0) + ascii_art(self._diagram, baseline=-1)

    def _unicode_art_(self):
        r"""
        Return a unicode art representation of ``self``.

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(QQ, 4)
            sage: SM = SGA.specht_module([(0,0), (1,1), (1,2), (2,1)])
            sage: unicode_art(SM)
             ┌─┬─┬─┐
             │X│ │ │
             ├─┼─┼─┤
             │ │X│X│
             ├─┼─┼─┤
             │ │X│ │
             └─┴─┴─┘
            S

            sage: SM = SGA.specht_module([3,1])
            sage: unicode_art(SM)
             ┌┬┬┐
             ├┼┴┘
             └┘
            S
        """
        from sage.typeset.unicode_art import unicode_art
        return unicode_art("S", baseline=0) + unicode_art(self._diagram, baseline=-1)

    class Element(SubmoduleWithBasis.Element):
        def _acted_upon_(self, x, self_on_left=False):
            """
            Return the action of ``x`` on ``self``.

            INPUT:

            - ``x`` -- an element of the base ring or can be converted into
              the defining symmetric group algebra
            - ``self_on_left`` -- boolean (default: ``False``); which side
              ``self`` is on for the action

            EXAMPLES::

                sage: SGA = SymmetricGroupAlgebra(QQ, 5)
                sage: from sage.combinat.diagram import Diagram
                sage: D = Diagram([(0,0), (1,1), (2,2), (3,3), (4,4)])
                sage: SM = SGA.specht_module(D)
                sage: SGA.an_element() * SM.an_element()
                15*S[0] + 6*S[1] + 9*S[2] + 6*S[3] + 6*S[4] + 2*S[72] + 2*S[96] + 3*S[97]

                sage: SGA = SymmetricGroupAlgebra(QQ, 4)
                sage: SM = SGA.specht_module([3,1])
                sage: SGA.an_element() * SM.an_element()
                9*S[[1, 2, 3], [4]] + 17*S[[1, 2, 4], [3]] + 14*S[[1, 3, 4], [2]]
                sage: 4 * SM.an_element()
                12*S[[1, 2, 3], [4]] + 8*S[[1, 2, 4], [3]] + 8*S[[1, 3, 4], [2]]

            TESTS::

                sage: SGA = SymmetricGroupAlgebra(QQ, 4)
                sage: SM = SGA.specht_module([3,1])
                sage: SM.an_element() * SGA.an_element()
                Traceback (most recent call last):
                ...
                TypeError: unsupported operand parent(s) for *:
                 'Specht module of [3, 1] over Rational Field'
                 and 'Symmetric group algebra of order 4 over Rational Field'
                sage: groups.permutation.Dihedral(3).an_element() * SM.an_element()
                Traceback (most recent call last):
                ...
                TypeError: unsupported operand parent(s) for *:
                 'Dihedral group of order 6 as a permutation group'
                 and 'Specht module of [3, 1] over Rational Field'
            """
            # Check for a scalar first
            ret = super()._acted_upon_(x, self_on_left)
            if ret is not None:
                return ret
            # Check if it is in the symmetric group algebra
            P = self.parent()
            if x in P._semigroup_algebra or x in P._semigroup_algebra.group():
                if self_on_left:  # it is only a left module
                    return None
                else:
                    return P.retract(P._semigroup_algebra(x) * self.lift())
            return None


class TabloidModule(SymmetricGroupRepresentation, CombinatorialFreeModule):
    r"""
    The vector space of all tabloids of a fixed shape with the natural
    symmetric group action.

    A *tabloid* is an :class:`OrderedSetPartition` whose underlying set
    is `\{1, \ldots, n\}`. The symmetric group acts by permuting the
    entries of the set. Hence, this is a representation of the symmetric
    group defined over any field.

    EXAMPLES::

        sage: SGA = SymmetricGroupAlgebra(GF(3), 5)
        sage: TM = SGA.tabloid_module([2, 2, 1])
        sage: TM.dimension()
        30
        sage: TM.brauer_character()
        (30, 6, 2, 0, 0)
        sage: IM = TM.invariant_module()
        sage: IM.dimension()
        1
        sage: IM.basis()[0].lift() == sum(TM.basis())
        True
    """
    @staticmethod
    def __classcall_private__(cls, SGA, shape):
        r"""
        Normalize input to ensure a unique representation.

        EXAMPLES::

            sage: from sage.combinat.specht_module import TabloidModule
            sage: SGA = SymmetricGroupAlgebra(QQ, 5)
            sage: TM1 = TabloidModule(SGA, [2, 2, 1])
            sage: TM2 = TabloidModule(SGA, Partition([2, 2, 1]))
            sage: TM1 is TM2
            True

            sage: TabloidModule(SGA, [3, 2, 1])
            Traceback (most recent call last):
            ...
            ValueError: the domain size (=5) does not match the number of boxes (=6) of the diagram
        """
        shape = _Partitions(shape)
        if SGA.group().rank() != sum(shape) - 1:
            rk = SGA.group().rank() + 1
            n = sum(shape)
            raise ValueError(f"the domain size (={rk}) does not match the number of boxes (={n}) of the diagram")
        return super().__classcall__(cls, SGA, shape)

    def __init__(self, SGA, shape):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(QQ, 5)
            sage: TM = SGA.tabloid_module([2,2,1])
            sage: TestSuite(TM).run()
        """
        from sage.combinat.set_partition_ordered import OrderedSetPartitions
        from sage.groups.perm_gps.permgroup_named import SymmetricGroup
        self._shape = shape
        n = sum(shape)
        self._symgp = SymmetricGroup(n)
        cat = ModulesWithBasis(SGA.base_ring()).FiniteDimensional()
        tabloids = OrderedSetPartitions(n, shape)
        CombinatorialFreeModule.__init__(self, SGA.base_ring(), tabloids,
                                         category=cat, prefix='T', bracket='')
        SymmetricGroupRepresentation.__init__(self, SGA)

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(QQ, 5)
            sage: SGA.tabloid_module([2,2,1])
            Tabloid module of [2, 2, 1] over Rational Field
        """
        return f"Tabloid module of {self._shape} over {self.base_ring()}"

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(QQ, 5)
            sage: latex(SGA.tabloid_module([2,2,1]))
            T^{{\def\lr#1{\multicolumn{1}{|@{\hspace{.6ex}}c@{\hspace{.6ex}}|}{\raisebox{-.3ex}{$#1$}}}
            \raisebox{-.6ex}{$\begin{array}[b]{*{2}c}\cline{1-2}
            \lr{\phantom{x}}&\lr{\phantom{x}}\\\cline{1-2}
            \lr{\phantom{x}}&\lr{\phantom{x}}\\\cline{1-2}
            \lr{\phantom{x}}\\\cline{1-1}
            \end{array}$}
            }}
        """
        from sage.misc.latex import latex
        return "T^{{{}}}".format(latex(self._shape))

    def _ascii_art_term(self, T):
        r"""
        Return an ascii art representation of the term indexed by ``T``.

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(QQ, 5)
            sage: TM = SGA.tabloid_module([2,2,1])
            sage: ascii_art(TM.an_element())  # indirect doctest
            2*T       + 2*T       + 3*T
               {1, 2}      {1, 2}      {1, 2}
               {3, 4}      {3, 5}      {4, 5}
               {5}         {4}         {3}
        """
        # This is basically copied from CombinatorialFreeModule._ascii_art_term
        from sage.typeset.ascii_art import AsciiArt, ascii_art
        pref = AsciiArt([self.prefix()])
        tab = "\n".join("{" + ", ".join(str(val) for val in sorted(row)) + "}" for row in T)
        if not tab:
            tab = '-'
        r = pref * (AsciiArt([" " * len(pref)]) + ascii_art(tab))
        r._baseline = r._h - 1
        return r

    def _unicode_art_term(self, T):
        r"""
        Return a unicode art representation of the term indexed by ``T``.

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(QQ, 5)
            sage: TM = SGA.tabloid_module([2,2,1])
            sage: unicode_art(TM.an_element())  # indirect doctest
            2*T       + 2*T       + 3*T
               {1, 2}      {1, 2}      {1, 2}
               {3, 4}      {3, 5}      {4, 5}
               {5}         {4}         {3}
        """
        from sage.typeset.unicode_art import unicode_art
        r = unicode_art(repr(self._ascii_art_term(T)))
        r._baseline = r._h - 1
        return r

    def _latex_term(self, T):
        r"""
        Return a latex representation of the term indexed by ``T``.

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(QQ, 5)
            sage: TM = SGA.tabloid_module([2,2,1])
            sage: latex(TM.an_element())  # indirect doctest
            2 T_{{\def\lr#1{\multicolumn{1}{@{\hspace{.6ex}}c@{\hspace{.6ex}}}{\raisebox{-.3ex}{$#1$}}}
            \raisebox{-.6ex}{$\begin{array}[b]{*{2}c}\cline{1-2}
            \lr{1}&\lr{2}\\\cline{1-2}
            \lr{3}&\lr{4}\\\cline{1-2}
            \lr{5}\\\cline{1-1}
            \end{array}$}
            }} + 2 T_{{\def\lr#1{\multicolumn{1}{@{\hspace{.6ex}}c@{\hspace{.6ex}}}{\raisebox{-.3ex}{$#1$}}}
            \raisebox{-.6ex}{$\begin{array}[b]{*{2}c}\cline{1-2}
            \lr{1}&\lr{2}\\\cline{1-2}
            \lr{3}&\lr{5}\\\cline{1-2}
            \lr{4}\\\cline{1-1}
            \end{array}$}
            }} + 3 T_{{\def\lr#1{\multicolumn{1}{@{\hspace{.6ex}}c@{\hspace{.6ex}}}{\raisebox{-.3ex}{$#1$}}}
            \raisebox{-.6ex}{$\begin{array}[b]{*{2}c}\cline{1-2}
            \lr{1}&\lr{2}\\\cline{1-2}
            \lr{4}&\lr{5}\\\cline{1-2}
            \lr{3}\\\cline{1-1}
            \end{array}$}
            }}
        """
        if not T:
            tab = "\\emptyset"
        else:
            from sage.combinat.output import tex_from_array
            A = list(map(sorted, T))
            tab = str(tex_from_array(A))
            tab = tab.replace("|", "")
        return f"{self.prefix()}_{{{tab}}}"

    def _symmetric_group_action(self, osp, g):
        r"""
        Return the action of the symmetric group element ``g`` on the
        ordered set partition ``osp``.

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(QQ, 5)
            sage: TM = SGA.tabloid_module([2,2,1])
            sage: osp = TM._indices([[1,4],[3,5],[2]])
            sage: g = SGA.group().an_element(); g
            [5, 1, 2, 3, 4]
            sage: TM._symmetric_group_action(osp, g)
            [{3, 5}, {2, 4}, {1}]
        """
        P = self._indices
        g = self._symgp(g)
        return P.element_class(P, [[g(val) for val in row] for row in osp], check=False)

    def specht_module(self):
        r"""
        Return the Specht submodule of ``self``.

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(QQ, 5)
            sage: TM = SGA.tabloid_module([2,2,1])
            sage: TM.specht_module() is SGA.specht_module([2,2,1])
            True
        """
        return SpechtModuleTableauxBasis(self)

    def bilinear_form(self, u, v):
        r"""
        Return the natural bilinear form of ``self`` applied to ``u`` and ``v``.

        The natural bilinear form is given by defining the tabloid basis
        to be orthonormal.

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(QQ, 5)
            sage: TM = SGA.tabloid_module([2,2,1])
            sage: u = TM.an_element(); u
            2*T[{1, 2}, {3, 4}, {5}] + 2*T[{1, 2}, {3, 5}, {4}] + 3*T[{1, 2}, {4, 5}, {3}]
            sage: v = sum(TM.basis())
            sage: TM.bilinear_form(u, v)
            7
            sage: TM.bilinear_form(u, TM.zero())
            0
        """
        if len(v) < len(u):
            u, v = v, u
        R = self.base_ring()
        return R.sum(c * v[T] for T, c in u)

    class Element(CombinatorialFreeModule.Element):
        def _acted_upon_(self, x, self_on_left):
            r"""
            Return the action of ``x`` on ``self``.

            INPUT:

            - ``x`` -- an element of the base ring or can be converted into
              the defining symmetric group algebra
            - ``self_on_left`` -- boolean (default: ``False``); which side
              ``self`` is on for the action

            EXAMPLES::

                sage: SGA = SymmetricGroupAlgebra(QQ, 5)
                sage: SM = SGA.tabloid_module([2,2,1])
                sage: SGA.an_element() * SM.an_element()
                2*T[{1, 5}, {2, 3}, {4}] + 2*T[{1, 5}, {2, 4}, {3}] + 3*T[{1, 5}, {3, 4}, {2}]
                 + 12*T[{1, 2}, {3, 4}, {5}] + 15*T[{1, 2}, {3, 5}, {4}] + 15*T[{1, 2}, {4, 5}, {3}]
                sage: 4 * SM.an_element()
                8*T[{1, 2}, {3, 4}, {5}] + 8*T[{1, 2}, {3, 5}, {4}] + 12*T[{1, 2}, {4, 5}, {3}]
                sage: SM.an_element() * SGA.an_element()
                Traceback (most recent call last):
                ...
                TypeError: unsupported operand parent(s) for *:
                 'Tabloid module of [2, 2, 1] over Rational Field'
                 and 'Symmetric group algebra of order 5 over Rational Field'
            """
            # first check for the base action
            ret = super()._acted_upon_(x, self_on_left)
            if ret is not None:
                return ret

            if self_on_left:
                return None
            P = self.parent()
            if x in P._semigroup_algebra:
                return P.linear_combination((perm * self, c) for perm, c in x.monomial_coefficients().items())
            if x in P._semigroup_algebra.indices():
                return P.element_class(P, {P._symmetric_group_action(T, x): c
                                           for T, c in self._monomial_coefficients.items()})


class SpechtModuleTableauxBasis(SpechtModule):
    r"""
    A Specht module of a partition in the classical standard
    tableau basis.

    This is constructed as a `S_n`-submodule of the :class:`TabloidModule`
    (also referred to as the standard module).

    .. SEEALSO::

        - :class:`SpechtModule` for the generic diagram implementation
          constructed as a left ideal of the group algebra
        - :class:`~sage.combinat.symmetric_group_representations.SpechtRepresentation`
          for an implementation of the representation by matrices.
    """
    def __init__(self, ambient):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(QQ, 5)
            sage: SM = SGA.specht_module([2,2,1])
            sage: TestSuite(SM).run()
        """
        self._diagram = ambient._shape
        SymmetricGroupRepresentation.__init__(self, ambient._semigroup_algebra)

        ambient_basis = ambient.basis()
        tabloids = ambient_basis.keys()
        support_order = list(tabloids)

        def elt(T):
            tab = tabloids.element_class(tabloids, list(T), check=False)
            return ambient.sum_of_terms((ambient._symmetric_group_action(tab, sigma), sigma.sign())
                                        for sigma in T.column_stabilizer())

        basis = Family({T: elt(T)
                        for T in self._diagram.standard_tableaux()})
        cat = ambient.category().Subobjects()
        SubmoduleWithBasis.__init__(self, basis, support_order, ambient=ambient,
                                    unitriangular=False, category=cat,
                                    prefix='S', bracket='')

    @lazy_attribute
    def lift(self):
        r"""
        The lift (embedding) map from ``self`` to the ambient space.

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(QQ, 5)
            sage: SM = SGA.specht_module([3, 1, 1])
            sage: SM.lift
            Generic morphism:
              From: Specht module of [3, 1, 1] over Rational Field
              To:   Tabloid module of [3, 1, 1] over Rational Field
        """
        return self.module_morphism(self.lift_on_basis, codomain=self.ambient())

    @lazy_attribute
    def retract(self):
        r"""
        The retract map from the ambient space.

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(QQ, 5)
            sage: X = SGA.tabloid_module([2,2,1])
            sage: Y = X.specht_module()
            sage: Y.retract
            Generic morphism:
              From: Tabloid module of [2, 2, 1] over Rational Field
              To:   Specht module of [2, 2, 1] over Rational Field
            sage: all(Y.retract(u.lift()) == u for u in Y.basis())
            True

            sage: Y.retract(X.zero())
            0
            sage: Y.retract(sum(X.basis()))
            Traceback (most recent call last):
            ...
            ValueError: ... is not in the image
        """
        B = self.basis()
        COB = matrix([b.lift().to_vector() for b in B]).T
        P, L, U = COB.LU()
        # Since U is upper triangular, the nonzero entries must be in the
        # upper square portion of the matrix
        n = len(B)

        Uinv = U.matrix_from_rows(range(n)).inverse()
        # This is a slight abuse as the codomain should be a module
        # with a different
        #  S_n action, but we only use it internally, so there is no problem
        PLinv = (P * L).inverse()

        def retraction(elt):
            vec = PLinv * elt.to_vector(order=self._support_order)
            if not vec:
                return self.zero()
            # vec is now in the image of self under U, which is
            if max(vec.support()) >= n:
                raise ValueError(f"{elt} is not in the image")
            return self._from_dict(dict(zip(B.keys(), Uinv * vec[:n])))

        return self._ambient.module_morphism(function=retraction, codomain=self)

    def bilinear_form(self, u, v):
        r"""
        Return the natural bilinear form of ``self`` applied to ``u`` and ``v``.

        The natural bilinear form is given by the pullback of the natural
        bilinear form on the tabloid module (where the tabloid basis is an
        orthonormal basis).

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(QQ, 5)
            sage: SM = SGA.specht_module([2,2,1])
            sage: u = SM.an_element(); u
            3*S[[1, 2], [3, 5], [4]] + 2*S[[1, 3], [2, 5], [4]] + 2*S[[1, 4], [2, 5], [3]]
            sage: v = sum(SM.basis())
            sage: SM.bilinear_form(u, v)
            140
        """
        TM = self._ambient
        return TM.bilinear_form(u.lift(), v.lift())

    @cached_method
    def gram_matrix(self):
        r"""
        Return the Gram matrix of the natural bilinear form of ``self``.

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(QQ, 5)
            sage: SM = SGA.specht_module([2,2,1])
            sage: M = SM.gram_matrix(); M
            [12  4 -4 -4  4]
            [ 4 12  4  4  4]
            [-4  4 12  4  4]
            [-4  4  4 12  4]
            [ 4  4  4  4 12]
            sage: M.det() != 0
            True
        """
        B = self.basis()
        M = matrix([[self.bilinear_form(b, bp) for bp in B] for b in B])
        M.set_immutable()
        return M

    def maximal_submodule(self):
        """
        Return the maximal submodule of ``self``.

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(GF(3), 5)
            sage: SM = SGA.specht_module([3,2])
            sage: U = SM.maximal_submodule()
            sage: U.dimension()
            4
        """
        return MaximalSpechtSubmodule(self)

    def simple_module(self):
        r"""
        Return the simple (or irreducible) `S_n`-submodule of ``self``.

        .. SEEALSO::

            :class:`~sage.combinat.specht_module.SimpleModule`

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(GF(3), 5)
            sage: SM = SGA.specht_module([3,2])
            sage: L = SM.simple_module()
            sage: L.dimension()
            1

            sage: SGA = SymmetricGroupAlgebra(QQ, 5)
            sage: SM = SGA.specht_module([3,2])
            sage: SM.simple_module() is SM
            True
        """
        if self.base_ring().characteristic() == 0:
            return self
        return SimpleModule(self)

    def intrinsic_arrangement(self, base_ring=None):
        r"""
        Return the intrinsic arrangement of ``self``.

        Consider the Specht module `S^{\lambda}` with `\lambda` a
        (integer) partition of `n` (i.e., `S^{\lambda}` is an `S_n`-module).
        The *intrinsic arrangement* of `S^{\lambda}` is the central hyperplane
        arrangement in `S^{\lambda}` given by the hyperplanes `H_{\alpha}`,
        indexed by a set partition `\alpha` of `\{1, \ldots, n\}` of size
        `\lambda`, defined by

        .. MATH::

            H_{\alpha} := \bigoplus_{\tau \in T_{\alpha}} (S^{\lambda})^{\tau},

        where `T_{\alpha}` is some set of generating transpositions
        of the Young subgroup `S_{\alpha}` and `V^{\tau}` denotes the
        `\tau`-invariant subspace of `V`. (These hyperplanes do not
        depend on the choice of `T_{\alpha}`.)

        This was introduced in [TVY2020]_ as a generalization of the
        braid arrangement, which is the case when `\lambda = (n-1, 1)`
        (equivalently, for the irreducible representation of `S_n`
        given by the type `A_{n-1}` root system).

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(QQ, 4)
            sage: SM = SGA.specht_module([2, 1, 1])
            sage: A = SM.intrinsic_arrangement()
            sage: A.hyperplanes()
            (Hyperplane T0 - T1 - 3*T2 + 0,
             Hyperplane T0 - T1 + T2 + 0,
             Hyperplane T0 + 3*T1 + T2 + 0,
             Hyperplane 3*T0 + T1 - T2 + 0)
            sage: A.is_free()
            False

        We reproduce Example 3 of [TVY2020]_::

            sage: SGA = SymmetricGroupAlgebra(QQ, 5)
            sage: for la in Partitions(5):
            ....:     SM = SGA.specht_module(la)
            ....:     A = SM.intrinsic_arrangement()
            ....:     print(la, A.characteristic_polynomial())
            [5] 1
            [4, 1] x^4 - 10*x^3 + 35*x^2 - 50*x + 24
            [3, 2] x^5 - 15*x^4 + 90*x^3 - 260*x^2 + 350*x - 166
            [3, 1, 1] x^6 - 10*x^5 + 45*x^4 - 115*x^3 + 175*x^2 - 147*x + 51
            [2, 2, 1] x^5 - 10*x^4 + 45*x^3 - 105*x^2 + 120*x - 51
            [2, 1, 1, 1] x^4 - 5*x^3 + 10*x^2 - 10*x + 4
            [1, 1, 1, 1, 1] 1

            sage: A = SGA.specht_module([4, 1]).intrinsic_arrangement()
            sage: A.characteristic_polynomial().factor()
            (x - 4) * (x - 3) * (x - 2) * (x - 1)
        """
        from sage.geometry.hyperplane_arrangement.arrangement import HyperplaneArrangements
        from sage.combinat.set_partition import SetPartitions
        if base_ring is None:
            base_ring = self.base_ring()

        if self.dimension() == 1:  # corner case
            HA = HyperplaneArrangements(base_ring, 'T')
            return HA()

        SGA = self._semigroup_algebra
        G = self._semigroup

        def t(i, j):
            ret = list(range(1, SGA.n + 1))
            ret[i - 1] = j
            ret[j - 1] = i
            return SGA(G(ret))

        # Construct the hyperplanes
        fixed_spaces = {}
        norms = []
        for alpha in SetPartitions(SGA.n, self._diagram.conjugate()):
            span = []
            for a in alpha:
                a = list(a)
                for i in range(len(a)-1):
                    elt = t(a[i], a[i+1])
                    if elt not in fixed_spaces:
                        fixed_spaces[elt] = self.annihilator_basis([elt - SGA.one()], side='left')
                    span.extend(fixed_spaces[elt])
            H = self.echelon_form(span)
            N = matrix([v.to_vector() for v in H]).right_kernel_matrix()
            assert N.nrows() == 1
            norms.append(N[0])

        # Convert the data to an arrangement
        HA = HyperplaneArrangements(base_ring, tuple([f'T{i}' for i in range(self.dimension())]))
        return HA([[0] + list(N) for N in norms])


class MaximalSpechtSubmodule(SymmetricGroupRepresentation, SubmoduleWithBasis):
    r"""
    The maximal submodule `U^{\lambda}` of the Specht module `S^{\lambda}`.

    ALGORITHM:

    We construct `U^{\lambda}` as the intersection `S \cap S^{\perp}`,
    where `S^{\perp}` is the orthogonal complement of the Specht module `S`
    inside of the tabloid module `T` (with respect to the natural
    bilinear form on `T`).

    EXAMPLES::

        sage: SGA = SymmetricGroupAlgebra(GF(3), 5)
        sage: SM = SGA.specht_module([3,2])
        sage: U = SM.maximal_submodule()
        sage: u = U.an_element(); u
        2*U[0] + 2*U[1]
        sage: [p * u for p in list(SGA.basis())[:4]]
        [2*U[0] + 2*U[1], 2*U[2] + 2*U[3], 2*U[0] + 2*U[1], U[0] + 2*U[2]]
        sage: sum(SGA.basis()) * u
        0
    """
    def __init__(self, specht_module):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(GF(3), 5)
            sage: SM = SGA.specht_module([3,2])
            sage: U = SM.maximal_submodule()
            sage: TestSuite(U).run()

            sage: SM = SGA.specht_module([2,1,1,1])
            sage: SM.maximal_submodule().dimension() == SM.dimension()
            True

            sage: SGA = SymmetricGroupAlgebra(QQ, 5)
            sage: SM = SGA.specht_module([3,2])
            sage: U = SM.maximal_submodule()
            sage: TestSuite(U).run()
            sage: U.dimension()
            0
        """
        SymmetricGroupRepresentation.__init__(self, specht_module._semigroup_algebra)

        p = specht_module.base_ring().characteristic()
        if p == 0:
            basis = Family([])
        else:
            TM = specht_module._ambient
            if not TM._shape.is_regular(p):
                basis = specht_module.basis()
            else:
                TV = TM._dense_free_module()
                SV = TV.submodule(specht_module.lift.matrix().columns())
                basis = (SV & SV.complement()).basis()
                basis = [specht_module.retract(TM.from_vector(b)) for b in basis]
                basis = Family(specht_module.echelon_form(basis))

        unitriangular = all(b.leading_support() == 1 for b in basis)
        support_order = list(specht_module.basis().keys())
        cat = specht_module.category().Subobjects()
        SubmoduleWithBasis.__init__(self, basis, support_order, ambient=specht_module,
                                    unitriangular=unitriangular, category=cat,
                                    prefix='U')

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(GF(3), 5)
            sage: SM = SGA.specht_module([3,2])
            sage: SM.maximal_submodule()
            Maximal submodule of Specht module of [3, 2] over Finite Field of size 3
        """
        return f"Maximal submodule of {self._ambient}"

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(GF(3), 5)
            sage: latex(SGA.specht_module([2,2,1]).maximal_submodule())
            U^{{\def\lr#1{\multicolumn{1}{|@{\hspace{.6ex}}c@{\hspace{.6ex}}|}{\raisebox{-.3ex}{$#1$}}}
            \raisebox{-.6ex}{$\begin{array}[b]{*{2}c}\cline{1-2}
            \lr{\phantom{x}}&\lr{\phantom{x}}\\\cline{1-2}
            \lr{\phantom{x}}&\lr{\phantom{x}}\\\cline{1-2}
            \lr{\phantom{x}}\\\cline{1-1}
            \end{array}$}
            }}
        """
        from sage.misc.latex import latex
        return "U^{{{}}}".format(latex(self._ambient._diagram))

    Element = SpechtModule.Element


class SimpleModule(SymmetricGroupRepresentation, QuotientModuleWithBasis):
    r"""
    The simple `S_n`-module associated with a partition `\lambda`.

    The simple module `D^{\lambda}` is the quotient of the Specht module
    `S^{\lambda}` by its :class:`maximal submodule <MaximalSpechtSubmodule>`
    `U^{\lambda}`.

    EXAMPLES::

        sage: SGA = SymmetricGroupAlgebra(GF(3), 5)
        sage: SM = SGA.specht_module([3,1,1])
        sage: D = SM.simple_module()
        sage: v = D.an_element(); v
        2*D[[[1, 3, 5], [2], [4]]] + 2*D[[[1, 4, 5], [2], [3]]]
        sage: SGA.an_element() * v
        2*D[[[1, 2, 4], [3], [5]]] + 2*D[[[1, 3, 5], [2], [4]]]

    We give an example on how to construct the decomposition matrix
    (the Specht modules are a complete set of irreducible projective
    modules) and the Cartan matrix of a symmetric group algebra::

        sage: SGA = SymmetricGroupAlgebra(GF(3), 4)
        sage: BM = matrix(SGA.simple_module(la).brauer_character()
        ....:             for la in Partitions(4, regular=3))
        sage: SBT = matrix(SGA.specht_module(la).brauer_character()
        ....:              for la in Partitions(4))
        sage: D = SBT * ~BM; D
        [1 0 0 0]
        [0 1 0 0]
        [1 0 1 0]
        [0 0 0 1]
        [0 0 1 0]
        sage: D.transpose() * D
        [2 0 1 0]
        [0 1 0 0]
        [1 0 2 0]
        [0 0 0 1]

    We verify this against the direct computation (up to reindexing the
    rows and columns)::

        sage: SGA.cartan_invariants_matrix()  # long time
        [1 0 0 0]
        [0 1 0 0]
        [0 0 2 1]
        [0 0 1 2]
    """
    def __init__(self, specht_module):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(GF(3), 5)
            sage: SM = SGA.specht_module([3,1,1])
            sage: D = SM.simple_module()
            sage: TestSuite(D).run()

            sage: SGA = SymmetricGroupAlgebra(GF(3), 5)
            sage: SM = SGA.specht_module([2,1,1,1])
            sage: SM.simple_module()
            Traceback (most recent call last):
            ...
            ValueError: the partition must be 3-regular
        """
        self._diagram = specht_module._diagram
        p = specht_module.base_ring().characteristic()
        if not self._diagram.is_regular(p):
            raise ValueError(f"the partition must be {p}-regular")
        SymmetricGroupRepresentation.__init__(self, specht_module._semigroup_algebra)
        cat = specht_module.category()
        QuotientModuleWithBasis.__init__(self, specht_module.maximal_submodule(), cat, prefix='D')

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(GF(3), 5)
            sage: SM = SGA.specht_module([3,1,1])
            sage: SM.simple_module()
            Simple module of [3, 1, 1] over Finite Field of size 3
        """
        return f"Simple module of {self._diagram} over {self.base_ring()}"

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: SGA = SymmetricGroupAlgebra(GF(3), 5)
            sage: latex(SGA.simple_module([2,2,1]))
            D^{{\def\lr#1{\multicolumn{1}{|@{\hspace{.6ex}}c@{\hspace{.6ex}}|}{\raisebox{-.3ex}{$#1$}}}
            \raisebox{-.6ex}{$\begin{array}[b]{*{2}c}\cline{1-2}
            \lr{\phantom{x}}&\lr{\phantom{x}}\\\cline{1-2}
            \lr{\phantom{x}}&\lr{\phantom{x}}\\\cline{1-2}
            \lr{\phantom{x}}\\\cline{1-1}
            \end{array}$}
            }}
        """
        from sage.misc.latex import latex
        return "D^{{{}}}".format(latex(self._diagram))

    Element = SpechtModule.Element


def _to_diagram(D):
    r"""
    Convert ``D`` to a list of cells representing a diagram.

    TESTS::

        sage: from sage.combinat.specht_module import _to_diagram
        sage: _to_diagram(Partition([3,1,1]))
        [(0, 0), (0, 1), (0, 2), (1, 0), (2, 0)]
        sage: _to_diagram(SkewPartition([[5,3,1,1],[2,2,1]]))
        [(0, 2), (0, 3), (0, 4), (1, 2), (3, 0)]
        sage: _to_diagram([1,2,0,2])
        [(0, 0), (1, 0), (1, 1), (3, 0), (3, 1)]
        sage: _to_diagram(Composition([2,1,3]))
        [(0, 0), (0, 1), (1, 0), (2, 0), (2, 1), (2, 2)]
        sage: _to_diagram([(1,2), (2,2)])
        [(1, 2), (2, 2)]
    """
    from sage.combinat.integer_vector import IntegerVectors
    from sage.combinat.skew_partition import SkewPartitions
    if isinstance(D, Diagram):
        return D
    if D in _Partitions:
        return _Partitions(D).cells()
    if D in SkewPartitions():
        return SkewPartitions()(D).cells()
    if D in IntegerVectors():
        return [(i, j) for i, row in enumerate(D) for j in range(row)]
    return [tuple(cell) for cell in D]


def specht_module_spanning_set(D, SGA=None):
    r"""
    Return a spanning set of the Specht module of diagram ``D``.

    INPUT:

    - ``D`` -- list of cells ``(r,c)`` for row ``r`` and column ``c``
    - ``SGA`` -- (optional) a symmetric group algebra

    EXAMPLES::

        sage: from sage.combinat.specht_module import specht_module_spanning_set
        sage: specht_module_spanning_set([(0,0), (1,1), (2,2)])
        ([1, 2, 3], [1, 3, 2], [2, 1, 3], [2, 3, 1], [3, 1, 2], [3, 2, 1])
        sage: specht_module_spanning_set([(0,0), (1,1), (2,1)])
        ([1, 2, 3] - [1, 3, 2], -[1, 2, 3] + [1, 3, 2], [2, 1, 3] - [3, 1, 2],
         [2, 3, 1] - [3, 2, 1], -[2, 1, 3] + [3, 1, 2], -[2, 3, 1] + [3, 2, 1])

        sage: SGA = SymmetricGroup(3).algebra(QQ)
        sage: specht_module_spanning_set([(0,0), (1,1), (2,1)], SGA)
        (() - (2,3), -(1,2) + (1,3,2), (1,2,3) - (1,3),
         -() + (2,3), -(1,2,3) + (1,3), (1,2) - (1,3,2))

    TESTS:

    Verify that diagrams bigger than the rank work::

        sage: specht_module_spanning_set([(0,0), (3,5)])
        ([1, 2], [2, 1])
        sage: specht_module_spanning_set([(0,0), (5,3)])
        ([1, 2], [2, 1])
    """
    n = len(D)
    if SGA is None:
        from sage.combinat.symmetric_group_algebra import SymmetricGroupAlgebra
        SGA = SymmetricGroupAlgebra(QQ, n)
    elif SGA.group().rank() != n - 1:
        raise ValueError("the rank does not match the size of the diagram")
    nr = max((c[0] for c in D), default=0) + 1
    nc = max((c[1] for c in D), default=0) + 1
    row_diagram = [set() for _ in range(nr)]
    col_diagram = [set() for _ in range(nc)]
    for i, cell in enumerate(D):
        x, y = cell
        row_diagram[x].add(i)
        col_diagram[y].add(i)
    # Construct the row and column stabilizer elements
    row_stab = SGA.zero()
    col_stab = SGA.zero()
    B = SGA.basis()
    for w in B.keys():
        # Remember that the permutation w is 1-based
        row_perm = [set() for _ in range(nr)]
        col_perm = [set() for _ in range(nc)]
        for i, cell in enumerate(D):
            x, y = cell
            row_perm[x].add(w(i + 1) - 1)
            col_perm[y].add(w(i + 1) - 1)
        if row_diagram == row_perm:
            row_stab += B[w]
        if col_diagram == col_perm:
            col_stab += w.sign() * B[w]
    gen = col_stab * row_stab
    return tuple([b * gen for b in B])


def specht_module_rank(D, base_ring=None):
    r"""
    Return the rank of the Specht module of diagram ``D``.

    EXAMPLES::

        sage: from sage.combinat.specht_module import specht_module_rank
        sage: specht_module_rank([(0,0), (1,1), (2,2)])
        6
    """
    D = _to_diagram(D)
    span_set = specht_module_spanning_set(D)
    if base_ring is None:
        base_ring = QQ
    return matrix(base_ring, [v.to_vector() for v in span_set]).rank()


def polytabloid(T):
    r"""
    Compute the polytabloid element associated to a tableau ``T``.

    For a tableau `T`, the polytabloid associated to `T` is

    .. MATH::

        e_T = \sum_{\sigma \in C_T} (-1)^{\sigma} \{\sigma T\},

    where `\{\}` is the row-equivalence class, i.e. a tabloid,
    and `C_T` is the column stabilizer of `T`. The sum takes place in
    the module spanned by tabloids `\{T\}`.

    OUTPUT:

    A ``dict`` whose keys are tabloids represented by tuples of frozensets
    and whose values are the coefficient.

    EXAMPLES::

        sage: from sage.combinat.specht_module import polytabloid
        sage: T = StandardTableau([[1,3,4],[2,5]])
        sage: polytabloid(T)
        {(frozenset({1, 3, 4}), frozenset({2, 5})): 1,
         (frozenset({1, 4, 5}), frozenset({2, 3})): -1,
         (frozenset({2, 3, 4}), frozenset({1, 5})): -1,
         (frozenset({2, 4, 5}), frozenset({1, 3})): 1}
    """
    e_T = {}
    C_T = T.column_stabilizer()
    for perm in C_T:
        TT = tuple([frozenset(perm(val) for val in row) for row in T])
        if TT in e_T:
            e_T[TT] += perm.sign()
        else:
            e_T[TT] = perm.sign()
    return e_T


def tabloid_gram_matrix(la, base_ring):
    r"""
    Compute the Gram matrix of the bilinear form of a Specht module
    pulled back from the tabloid module.

    For the module spanned by all tabloids, we define an bilinear form
    by having the tabloids be an orthonormal basis. We then pull this
    bilinear form back across the natural injection of the Specht module
    into the tabloid module.

    EXAMPLES::

        sage: from sage.combinat.specht_module import tabloid_gram_matrix
        sage: tabloid_gram_matrix([3,2], GF(5))
        [4 2 2 1 4]
        [2 4 1 2 1]
        [2 1 4 2 1]
        [1 2 2 4 2]
        [4 1 1 2 4]
    """
    from sage.combinat.tableau import StandardTableaux
    ST = list(StandardTableaux(la))

    def bilinear_form(p1, p2):
        if len(p2) < len(p1):
            p1, p2 = p2, p1
        return sum(c1 * p2[T1] for T1, c1 in p1.items() if c1 and T1 in p2)

    PT = {T: polytabloid(T) for T in ST}
    gram_matrix = [[bilinear_form(PT[T1], PT[T2]) for T1 in ST] for T2 in ST]
    return matrix(base_ring, gram_matrix)


def simple_module_rank(la, base_ring):
    r"""
    Return the rank of the simple `S_n`-module corresponding to the
    partition ``la`` of size `n` over ``base_ring``.

    EXAMPLES::

        sage: from sage.combinat.specht_module import simple_module_rank
        sage: simple_module_rank([3,2,1,1], GF(3))
        13

    TESTS::

        sage: from sage.combinat.specht_module import simple_module_rank
        sage: simple_module_rank([1,1,1,1], GF(3))
        Traceback (most recent call last):
        ...
        ValueError: the partition [1, 1, 1, 1] is not 3-regular

        sage: from sage.combinat.specht_module import simple_module_rank
        sage: simple_module_rank([2,1], GF(3)['x'])
        Traceback (most recent call last):
        ...
        NotImplementedError: the base must be a field
    """
    from sage.categories.fields import Fields
    from sage.combinat.partition import Partition
    if base_ring not in Fields():
        raise NotImplementedError("the base must be a field")
    p = base_ring.characteristic()
    la = Partition(la)
    if not la.is_regular(p):
        raise ValueError(f"the partition {la} is not {p}-regular")
    return tabloid_gram_matrix(la, base_ring).rank()



# sage_setup: distribution = sagemath-combinat
r"""
Degree sequences

The present module implements the ``DegreeSequences`` class, whose instances
represent the integer sequences of length `n`::

    sage: DegreeSequences(6)
    Degree sequences on 6 elements

With the object ``DegreeSequences(n)``, one can:

* Check whether a sequence is indeed a degree sequence::

    sage: DS = DegreeSequences(5)
    sage: [4, 3, 3, 3, 3] in DS
    True
    sage: [4, 4, 0, 0, 0] in DS
    False

* List all the possible degree sequences of length `n`::

    sage: for seq in DegreeSequences(4):
    ....:     print(seq)
    [0, 0, 0, 0]
    [1, 1, 0, 0]
    [2, 1, 1, 0]
    [3, 1, 1, 1]
    [1, 1, 1, 1]
    [2, 2, 1, 1]
    [2, 2, 2, 0]
    [3, 2, 2, 1]
    [2, 2, 2, 2]
    [3, 3, 2, 2]
    [3, 3, 3, 3]

.. NOTE::

    Given a degree sequence, one can obtain a graph realizing it by using
    :func:`~sage.graphs.generators.degree_sequence.DegreeSequence`.
    For instance::

        sage: ds = [3, 3, 2, 2, 2, 2, 2, 1, 1, 0]
        sage: g = graphs.DegreeSequence(ds)                                             # needs networkx sage.graphs
        sage: g.degree_sequence()                                                       # needs networkx sage.graphs
        [3, 3, 2, 2, 2, 2, 2, 1, 1, 0]

Definitions
~~~~~~~~~~~

A sequence of integers `d_1,...,d_n` is said to be a *degree sequence* (or
*graphic* sequence) if there exists a graph in which vertex `i` is of degree
`d_i`. It is often required to be *non-increasing*, i.e. that
`d_1 \geq ... \geq d_n`. Finding a graph with given degree sequence is
known as *graph realization problem*.

An integer sequence need not necessarily be a degree sequence. Indeed, in a
degree sequence of length `n` no integer can be larger than `n-1` -- the degree
of a vertex is at most `n-1` -- and the sum of them is at most `n(n-1)`.

Degree sequences are completely characterized by a result from Erdős and Gallai:

**Erdős and Gallai:** *The sequence of integers* `d_1 \geq \cdots \geq d_n`
*is a degree sequence if and only if* `\sum_i d_i` is even and `\forall i`

.. MATH::

    \sum_{j\leq i}d_j \leq j(j-1) + \sum_{j>i} \min(d_j,i).

Alternatively, a degree sequence can be defined recursively:

**Havel and Hakimi:** *The sequence of integers* `d_1\geq ... \geq d_n` *is a
degree sequence if and only if* `d_2-1,...,d_{d_1+1}-1, d_{d_1+2}, ...,d_n` *is
also a degree sequence.*

Or equivalently:

**Havel and Hakimi (bis):** *If there is a realization of an integer sequence as
a graph (i.e. if the sequence is a degree sequence), then it can be realized in
such a way that the vertex of maximum degree* `\Delta` *is adjacent to the*
`\Delta` *vertices of highest degree (except itself, of course).*


Algorithms
~~~~~~~~~~

**Checking whether a given sequence is a degree sequence**

This is tested using Erdos and Gallai's criterion. It is also checked that the
given sequence is non-increasing and has length `n`.

**Iterating through the sequences of length** `n`

From Havel and Hakimi's recursive definition of a degree sequence, one can
build an enumeration algorithm as done in [RCES1994]_. It consists in
trying to **extend** a current degree sequence on `n` elements into a
degree sequence on `n+1` elements by adding a vertex of degree larger
than those already present in the sequence. This can be seen as **reversing**
the reduction operation described in Havel and Hakimi's characterization.
This operation can appear in several different ways:

* Extensions of a degree sequence that do **not** change the value of the
  maximum element

  * If the maximum element of a given degree sequence is `0`, then one can
    remove it to reduce the sequence, following Havel and Hakimi's
    rule. Conversely, if the maximum element of the (current) sequence is
    `0`, then one can always extend it by adding a new element of degree
    `0` to the sequence.

    .. MATH::

        0, 0, 0 \xrightarrow{Extension} {\bf 0}, 0, 0, 0 \xrightarrow{Extension}
        {\bf 0}, 0, 0, ..., 0, 0, 0 \xrightarrow{Reduction} 0, 0, 0, 0
        \xrightarrow{Reduction} 0, 0, 0

  * If there are at least `\Delta+1` elements of (maximum) degree `\Delta`
    in a given degree sequence, then one can reduce it by removing a
    vertex of degree `\Delta` and decreasing the values of `\Delta`
    elements of value `\Delta` to `\Delta-1`. Conversely, if the maximum
    element of the (current) sequence is `d>0`, then one can add a new
    element of degree `d` to the sequence if it can be linked to `d`
    elements of (current) degree `d-1`. Those `d` vertices of degree `d-1`
    hence become vertices of degree `d`, and so `d` elements of degree
    `d-1` are removed from the sequence while `d+1` elements of degree `d`
    are added to it.

    .. MATH::

        3, 2, 2, 2, 1 \xrightarrow{Extension} {\bf 3}, 3, (2+1), (2+1), (2+1), 1
        = {\bf 3}, 3, 3, 3, 3, 1 \xrightarrow{Reduction} 3, 2, 2, 2, 1

* Extension of a degree sequence that changes the value of the maximum
  element:

  * In the general case, i.e. when the number of elements of value
    `\Delta,\Delta-1` is small compared to `\Delta` (i.e. the maximum
    element of a given degree sequence), reducing a sequence strictly
    decreases the value of the maximum element. According to Havel and
    Hakimi's characterization there is only **one** way to reduce a
    sequence, but reversing this operation is more complicated than in the
    previous cases. Indeed, the following extensions are perfectly valid
    according to the reduction rule.

    .. MATH::

        2,1,1,0,0\xrightarrow{Extension} {\bf 3}, (2+1), (1+1), (1+1), 0, 0
        = 3, 3, 2, 2, 0, 0 \xrightarrow{Reduction} 2, 1, 1, 0, 0\\
        2,1,1,0,0\xrightarrow{Extension} {\bf 3}, (2+1), (1+1), 1, (0+1), 0
        = 3, 3, 2, 1, 1, 0 \xrightarrow{Reduction} 2, 1, 1, 0, 0\\
        2,1,1,0,0\xrightarrow{Extension} {\bf 3}, (2+1), 1, 1, (0+1), (0+1)
        = 3, 3, 1, 1, 1, 1 \xrightarrow{Reduction} 2, 1, 1, 0, 0\\
        ...

    In order to extend a current degree sequence while strictly increasing
    its maximum degree, it is equivalent to pick a set `I` of elements of
    the degree sequence with `|I|>\Delta` in such a way that the
    `(d_i+1)_{i\in I}` are the `|I|` maximum elements of the sequence
    `(d_i+\genfrac{}{}{0pt}{}{1\text{ if }i\in I}{0\text{ if }i\not \in
    I})_{1\leq i \leq n}`, and to add to this new sequence an element of
    value `|I|`. The non-increasing sequence containing the elements `|I|`
    and `(d_i+\genfrac{}{}{0pt}{}{1\text{ if }i\in I}{0\text{ if }i\not
    \in I})_{1\leq i \leq n}` can be reduced to `(d_i)_{1\leq i \leq n}`
    by Havel and Hakimi's rule.

    .. MATH::

        ... 1, 1, 2, {\bf 2}, {\bf 2}, 2, 2, 3, 3, \underline{3}, {\bf 3},
        {\bf 3}, {\bf 4}, {\bf 6}, ... \xrightarrow{Extension} ... 1, 1,
        2, 2, 2, 3, 3, \underline{3}, {\bf 3}, {\bf 3}, {\bf 4}, {\bf 4},
        {\bf 5}, {\bf 7}, ...

    The number of possible sets `I` having this property (i.e. the number
    of possible extensions of a sequence) is smaller than it
    seems. Indeed, by definition, if `j\not \in I` then for all `i\in I`
    the inequality `d_j\leq d_i+1` holds. Hence, each set `I` is entirely
    determined by the largest element `d_k` of the sequence that it does
    **not** contain (hence `I` contains `\{1,...,k-1\}`), and by the
    cardinalities of `\{i\in I:d_i= d_k\}` and `\{i\in I:d_i= d_k-1\}`.

    .. MATH::

        I = \{i \in I : d_i= d_k \} \cup \{i \in I : d_i= d_k-1 \}
        \cup \{i : d_i> d_k \}.

    The number of possible extensions is hence at most cubic, and is
    easily enumerated.

About the implementation
~~~~~~~~~~~~~~~~~~~~~~~~

In the actual implementation of the enumeration algorithm, the degree sequence
is stored differently for reasons of efficiency.

Indeed, when enumerating all the degree sequences of length `n`, Sage first
allocates an array ``seq`` of `n+1` integers where ``seq[i]`` is the number of
elements of value ``i`` in the current sequence. Obviously, ``seq[n]=0`` holds
in permanence : it is useful to allocate a larger array than necessary to
simplify the code. The ``seq`` array is a global variable.

The recursive function ``enum(depth, maximum)`` is the one building the list of
sequences. It builds the list of degree sequences of length `n` which *extend*
the sequence currently stored in ``seq[0]...seq[depth-1]``. When it is called,
``maximum`` must be set to the maximum value of an element in the partial
sequence ``seq[0]...seq[depth-1]``.

If during its run the function ``enum`` heavily works on the content of
the ``seq`` array, the value of ``seq`` is the **same** before and after
the run of ``enum``.

**Extending the current partial sequence**

The two cases for which the maximum degree of the partial sequence does not
change are easy to detect. It is (slightly) harder to enumerate all the sets
`I` corresponding to possible extensions of the partial sequence. As said
previously, to each set `I` one can associate an integer ``current_box`` such
that `I` contains all the `i` satisfying `d_i>current\_box`. The variable
``taken`` represents the number of all such elements `i`, so that when
enumerating all possible sets `I` in the algorithm we have the equality

.. MATH::

    I = \text{taken }+\text{ number of elements of value }current\_box+
    \text{ number of elements of value }current\_box-1.

REFERENCES:

- [RCES1994]_

AUTHORS:

- Nathann Cohen

TESTS:

The sequences produced by random graphs *are* degree sequences::

    sage: n = 30
    sage: DS = DegreeSequences(30)
    sage: for i in range(10):                                                           # needs networkx sage.graphs
    ....:     g = graphs.RandomGNP(n,.2)
    ....:     if not g.degree_sequence() in DS:
    ....:         print("Something is very wrong !")

Checking that we indeed enumerate *all* the degree sequences for `n=5`::

    sage: ds1 = Set([tuple(g.degree_sequence()) for g in graphs(5)])                    # needs nauty sage.graphs
    sage: ds2 = Set(map(tuple,list(DegreeSequences(5))))
    sage: ds1 == ds2                                                                    # needs nauty sage.graphs
    True

Checking the consistency of enumeration and test::

    sage: DS = DegreeSequences(6)
    sage: all(seq in DS for seq in DS)
    True

.. WARNING::

    For the moment, iterating over all degree sequences involves building the
    list of them first, then iterate on this list.  This is obviously bad,
    as it requires uselessly a **lot** of memory for large values of `n`.

    This should be changed. Updating the code does not require more
    than a couple of minutes.
"""

# ****************************************************************************
#       Copyright (C) 2011 Nathann Cohen <nathann.cohen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from cysignals.memory cimport check_calloc, sig_free
from cysignals.signals cimport sig_on, sig_off


cdef unsigned char * seq
cdef list sequences


class DegreeSequences:

    def __init__(self, n):
        r"""
        Degree Sequences.

        An instance of this class represents the degree sequences of graphs on a
        given number `n` of vertices. It can be used to list and count them, as
        well as to test whether a sequence is a degree sequence. For more
        information, please refer to the documentation of the
        :mod:`DegreeSequence<sage.combinat.degree_sequences>` module.

        EXAMPLES::

            sage: DegreeSequences(8)
            Degree sequences on 8 elements
            sage: [3,3,2,2,2,2,2,2] in DegreeSequences(8)
            True

        TESTS:

        :issue:`21824`::

            sage: DegreeSequences(-1)
            Traceback (most recent call last):
            ...
            ValueError: the input parameter must be >= 0
        """
        if n < 0:
            raise ValueError("the input parameter must be >= 0")
        self._n = n

    def __contains__(self, seq):
        """
        Check whether a given integer sequence is the degree sequence
        of a graph on `n` elements.

        EXAMPLES::

            sage: [3,3,2,2,2,2,2,2] in DegreeSequences(8)
            True

        TESTS:

        :issue:`15503`::

            sage: [2,2,2,2,1,1,1] in DegreeSequences(7)
            False

        :issue:`21824`::

            sage: [d for d in DegreeSequences(0)]
            [[]]
            sage: [d for d in DegreeSequences(1)]
            [[0]]
            sage: [d for d in DegreeSequences(3)]
            [[0, 0, 0], [1, 1, 0], [2, 1, 1], [2, 2, 2]]
            sage: [d for d in DegreeSequences(1)]
            [[0]]
        """
        cdef int n = self._n
        if len(seq) != n:
            return False

        # Is the sum even ?
        if sum(seq) % 2:
            return False

        # Partial represents the left side of Erdos and Gallai's inequality,
        # i.e. the sum of the i first integers.
        cdef int partial = 0
        cdef int i, d, dd, right

        # Temporary variable to ensure that the sequence is indeed
        # non-increasing
        cdef int prev = n - 1

        for i, d in enumerate(seq):

            # Non-increasing ?
            if d > prev:
                return False
            else:
                prev = d

            # Updating the partial sum
            partial += d

            # Evaluating the right hand side
            right = i * (i + 1)
            for dd in seq[i + 1:]:
                right += min(dd, i + 1)

            # Comparing the two
            if partial > right:
                return False

        return True

    def __repr__(self):
        """
        Representing the element.

        TESTS::

            sage: DegreeSequences(6)
            Degree sequences on 6 elements
        """
        return "Degree sequences on "+str(self._n)+" elements"

    def __iter__(self):
        """
        Iterate over all the degree sequences.

        TODO: THIS SHOULD BE UPDATED AS SOON AS THE YIELD KEYWORD APPEARS IN
        CYTHON. See comment in the class' documentation.

        EXAMPLES::

            sage: DS = DegreeSequences(6)
            sage: all(seq in DS for seq in DS)
            True
        """
        yield from init(self._n)

    def __dealloc__():
        """
        Freeing the memory
        """
        sig_free(seq)


cdef init(int n):
    """
    Initialize the memory and starts the enumeration algorithm.
    """
    global seq
    global N
    global sequences

    if n == 0:
        return [[]]
    elif n == 1:
        return [[0]]

    seq = <unsigned char *>check_calloc(n + 1, sizeof(unsigned char))

    # We begin with one vertex of degree 0
    seq[0] = 1

    N = n
    sequences = []
    enum(1, 0)
    sig_free(seq)
    return sequences

cdef inline add_seq():
    """
    This function is called whenever a sequence is found.

    Build the degree sequence corresponding to the current state of the
    algorithm and adds it to the sequences list.
    """
    global sequences
    global N
    global seq

    cdef list s = []
    cdef int i, j

    for N > i >= 0:
        for 0 <= j < seq[i]:
            s.append(i)

    sequences.append(s)


cdef void enum(int k, int M) noexcept:
    r"""
    Main function; for an explanation of the algorithm please refer to the
    :mod:`sage.combinat.degree_sequences` documentation.

    INPUT:

    - ``k`` -- depth of the partial degree sequence
    - ``M`` -- value of a maximum element in the partial degree sequence
    """
    cdef int i, j
    global seq
    cdef int taken = 0
    cdef int current_box
    cdef int n_current_box
    cdef int n_previous_box
    cdef int new_vertex

    # Have we found a new degree sequence ? End of recursion !
    if k == N:
        add_seq()
        return

    sig_on()

    #############################################
    # Creating vertices of Vertices of degree M #
    #############################################

    # If 0 is the current maximum degree, we can always extend the degree
    # sequence with another 0
    if M == 0:

        seq[0] += 1
        enum(k + 1, M)
        seq[0] -= 1

    # We need not automatically increase the degree at each step. In this case,
    # we have no other choice but to link the new vertex of degree M to vertices
    # of degree M-1, which will become vertices of degree M too.
    elif seq[M - 1] >= M:

        seq[M] += M + 1
        seq[M - 1] -= M

        enum(k + 1, M)

        seq[M] -= M + 1
        seq[M - 1] += M

    ###############################################
    # Creating vertices of Vertices of degree > M #
    ###############################################

    for M >= current_box > 0:

        # If there is not enough vertices in the boxes available
        if taken + (seq[current_box] - 1) + seq[current_box-1] <= M:
            taken += seq[current_box]
            seq[current_box+1] += seq[current_box]
            seq[current_box] = 0
            continue

        # The degree of the new vertex will be taken + i + j where:
        #
        # * i is the number of vertices taken in the *current* box
        # * j the number of vertices taken in the *previous* one

        n_current_box = seq[current_box]
        n_previous_box = seq[current_box-1]

        # Note to self, and others:
        #
        # In the following lines, there are many incrementation/decrementation
        # that *may* be replaced by only +1 and -1 and save some
        # instructions. This would involve adding several "if", and I feared it
        # would make the code even uglier. If you are willing to give it a try,
        # **please check the results** ! It is trickier that it seems ! Even
        # changing the lower bounds in the for loops would require tests
        # afterwards.

        for max(0, (M+1)-n_previous_box-taken) <= i < n_current_box:
            seq[current_box] -= i
            seq[current_box+1] += i

            for max(0, ((M+1)-taken-i)) <= j <= n_previous_box:
                seq[current_box-1] -= j
                seq[current_box] += j

                new_vertex = taken + i + j
                seq[new_vertex] += 1
                enum(k+1, new_vertex)
                seq[new_vertex] -= 1

                seq[current_box-1] += j
                seq[current_box] -= j

            seq[current_box] += i
            seq[current_box+1] -= i

        taken += n_current_box
        seq[current_box] = 0
        seq[current_box+1] += n_current_box

    # Corner case
    #
    # Now current_box = 0. All the vertices of nonzero degree are taken, we just
    # want to know how many vertices of degree 0 will be neighbors of the new
    # vertex.
    for max(0, ((M+1)-taken)) <= i <= seq[0]:

        seq[1] += i
        seq[0] -= i
        seq[taken+i] += 1

        enum(k+1, taken+i)

        seq[taken+i] -= 1
        seq[1] -= i
        seq[0] += i

    # Shift everything back to normal ! ( cell N is always equal to 0)
    for 1 <= i < N:
        seq[i] = seq[i+1]

    sig_off()



# sage_setup: distribution = sagemath-combinat
r"""
Rigged configurations

.. TODO:: Proofread / point to the main classes rather than the modules?

- :ref:`sage.combinat.rigged_configurations.rc_crystal`
- :ref:`sage.combinat.rigged_configurations.rc_infinity`

- :ref:`sage.combinat.rigged_configurations.rigged_configurations`
- :ref:`sage.combinat.rigged_configurations.rigged_configuration_element`

- :ref:`sage.combinat.rigged_configurations.tensor_product_kr_tableaux`
- :ref:`sage.combinat.rigged_configurations.tensor_product_kr_tableaux_element`
- :ref:`sage.combinat.rigged_configurations.kr_tableaux`

- :ref:`sage.combinat.rigged_configurations.kleber_tree`

- :ref:`sage.combinat.rigged_configurations.rigged_partition`

Bijections
----------

- :ref:`sage.combinat.rigged_configurations.bijection`
- :ref:`sage.combinat.rigged_configurations.bij_abstract_class`
- :ref:`sage.combinat.rigged_configurations.bij_type_A`
- :ref:`sage.combinat.rigged_configurations.bij_type_B`
- :ref:`sage.combinat.rigged_configurations.bij_type_C`
- :ref:`sage.combinat.rigged_configurations.bij_type_D`
- :ref:`sage.combinat.rigged_configurations.bij_type_A2_odd`
- :ref:`sage.combinat.rigged_configurations.bij_type_A2_even`
- :ref:`sage.combinat.rigged_configurations.bij_type_A2_dual`
- :ref:`sage.combinat.rigged_configurations.bij_type_D_twisted`
- :ref:`sage.combinat.rigged_configurations.bij_type_D_tri`
- :ref:`sage.combinat.rigged_configurations.bij_infinity`
"""
# install the docstring of this module to the containing package
from sage.misc.namespace_package import install_doc
install_doc(__package__, __doc__)

from sage.misc.lazy_import import lazy_import

lazy_import('sage.combinat.rigged_configurations.rigged_configurations',
            'RiggedConfigurations')
del lazy_import
del install_doc



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.groups sage.modules
"""
Yokonuma-Hecke Algebras

AUTHORS:

- Travis Scrimshaw (2015-11): initial version
- Travis Scrimshaw (2025-03): general type version
"""

# ****************************************************************************
#  Copyright (C) 2015-2025 Travis Scrimshaw <tcscrims at gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.cachefunc import cached_method
from sage.rings.integer_ring import ZZ
from sage.rings.polynomial.laurent_polynomial_ring import LaurentPolynomialRing
from sage.rings.rational_field import QQ
from sage.categories.algebras import Algebras
from sage.categories.rings import Rings
from sage.combinat.free_module import CombinatorialFreeModule
from sage.combinat.permutation import Permutations
from sage.sets.family import Family


class YokonumaHeckeAlgebra(CombinatorialFreeModule):
    r"""
    Abstract base class for Yokonuma-Hecke algebras that
    implements common features.

    .. TODO::

        Factor out the near-common features.
    """
    @staticmethod
    def __classcall_private__(cls, d, n, q=None, R=None):
        r"""
        Standardize input to ensure a unique representation and dispatch
        to the correct implementation.

        TESTS::

            sage: Y1 = algebras.YokonumaHecke(5, 3)
            sage: q = LaurentPolynomialRing(QQ, 'q').gen()
            sage: Y2 = algebras.YokonumaHecke(5, 3, q)
            sage: Y3 = algebras.YokonumaHecke(5, 3, q, q.parent())
            sage: Y1 is Y2 and Y2 is Y3
            True
        """
        if q is None:
            q = LaurentPolynomialRing(QQ, 'q').gen()
        if R is None:
            R = q.parent()
        q = R(q)
        if R not in Rings().Commutative():
            raise TypeError("base ring must be a commutative ring")
        if n not in ZZ:
            from sage.combinat.root_system.cartan_type import CartanType
            n = CartanType(n)
            return YokonumaHeckeAlgebraWeyl(d, n, q, R)
        return YokonumaHeckeAlgebraGL(d, n, q, R)

    def __init__(self, d, W, q, R, indices, category=None):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(5, 3)
            sage: elts = Y.some_elements() + list(Y.algebra_generators())
            sage: TestSuite(Y).run(elements=elts)
        """
        self._d = d
        self._W = W
        self._cartan_type = W.cartan_type()
        self._q = q
        cat = Algebras(R).WithBasis().or_subcategory(category)
        CombinatorialFreeModule.__init__(self, R, indices, prefix='Y',
                                         category=cat)
        self._assign_names(self.algebra_generators().keys())

    def cartan_type(self):
        r"""
        Return the Cartan type of ``self``.

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(4, ['F',4])
            sage: Y.cartan_type()
            ['F', 4]
        """
        return self._cartan_type

    def index_set(self):
        r"""
        Return the index set of ``self``, which is the index set of
        the Cartan type of ``self``.

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(4, ['F',4])
            sage: Y.index_set() == Y.cartan_type().index_set()
            True
        """
        return self._cartan_type.index_set()

    def q(self):
        r"""
        Return the parameter `q` of ``self``.

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(4, ['F',4])
            sage: Y.q()
            q
            sage: Y.q().parent() is Y.base_ring()
            True
        """
        return self._q

    def g(self, i=None):
        """
        Return the generator(s) `g_i`.

        INPUT:

        - ``i`` -- (default: ``None``) the generator `g_i` or if ``None``,
          then the family of all generators `g_i`

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(8, 3)
            sage: Y.g(1)
            g[1]
            sage: list(Y.g())
            [g[1], g[2]]

            sage: Y = algebras.YokonumaHecke(8, ['G',2])
            sage: Y.g(1)
            g[1]
            sage: Y.g()
            Finite family {1: g[1], 2: g[2]}
        """
        G = self.algebra_generators()
        if i is None:
            I = self._W.index_set()
            d = {i: G['g%s' % i] for i in I}
            return Family(I, d.__getitem__)
        return G['g%s' % i]

    @cached_method
    def gens(self) -> tuple:
        """
        Return the generators of ``self``.

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(5, 3)
            sage: Y.gens()
            (g[1], g[2], t1, t2, t3)

            sage: Y = algebras.YokonumaHecke(5, ['B',2])
            sage: Y.gens()
            (g[1], g[2], h1, h2)
        """
        return tuple(self.algebra_generators())


class YokonumaHeckeAlgebraGL(YokonumaHeckeAlgebra):
    r"""
    The Yokonuma-Hecke algebra `Y_{d,n}(q)` for `GL_n(\GF{d})`.

    Let `R` be a commutative ring and `q` be a unit in `R`. The
    *Yokonuma-Hecke algebra* `Y_{d,n}(q)` is the associative, unital
    `R`-algebra generated by `t_1, t_2, \ldots, t_n, g_1, g_2, \ldots,
    g_{n-1}` and subject to the relations:

    - `g_i g_j = g_j g_i` for all `|i - j| > 1`,
    - `g_i g_{i+1} g_i = g_{i+1} g_i g_{i+1}`,
    - `t_i t_j = t_j t_i`,
    - `t_j g_i = g_i t_{j s_i}`, and
    - `t_j^d = 1`,

    where `s_i` is the simple transposition `(i, i+1)`, along with
    the quadratic relation

    .. MATH::

        g_i^2 = 1 + \frac{(q - q^{-1})}{d} \left( \sum_{s=0}^{d-1}
        t_i^s t_{i+1}^{-s} \right) g_i.

    Thus the Yokonuma-Hecke algebra can be considered a quotient of
    the framed braid group `(\ZZ / d\ZZ) \wr B_n`, where `B_n` is the
    classical braid group on `n` strands, by the quadratic relations.
    Moreover, all of the algebra generators are invertible. In
    particular, we have

    .. MATH::

        g_i^{-1} = g_i - (q - q^{-1}) e_i.

    When we specialize `q = \pm 1`, we obtain the group algebra of
    the complex reflection group `G(d, 1, n) = (\ZZ / d\ZZ) \wr S_n`.
    Moreover for `d = 1`, the Yokonuma-Hecke algebra is equal to the
    :class:`Iwahori-Hecke <IwahoriHeckeAlgebra>` of type `A_{n-1}`.

    This was considered for more general Chevalley groups (Lie groups
    over finite fields); see :class:`YokonumaHeckeAlgebraWeyl`.

    INPUT:

    - ``d`` -- the maximum power of `t`
    - ``n`` -- the number of generators or a Cartan type
    - ``q`` -- (optional) an invertible element in a commutative ring;
      the default is `q \in \QQ[q,q^{-1}]`
    - ``R`` -- (optional) a commutative ring containing ``q``; the
      default is the parent of `q`

    EXAMPLES:

    We construct `Y_{4,3}` and do some computations::

        sage: Y = algebras.YokonumaHecke(4, 3)
        sage: g1, g2, t1, t2, t3 = Y.algebra_generators()
        sage: g1 * g2
        g[1,2]
        sage: t1 * g1
        t1*g[1]
        sage: g2 * t2
        t3*g[2]
        sage: g2 * t3
        t2*g[2]
        sage: (g2 + t1) * (g1 + t2*t3)
        g[2,1] + t2*t3*g[2] + t1*g[1] + t1*t2*t3
        sage: g1 * g1
        1 - (1/4*q^-1-1/4*q)*g[1] - (1/4*q^-1-1/4*q)*t1*t2^3*g[1]
         - (1/4*q^-1-1/4*q)*t1^2*t2^2*g[1] - (1/4*q^-1-1/4*q)*t1^3*t2*g[1]
        sage: g2 * g1 * t1
        t3*g[2,1]

    We construct the elements `e_i` and show that they are idempotents::

        sage: e1 = Y.e(1); e1
        1/4 + 1/4*t1*t2^3 + 1/4*t1^2*t2^2 + 1/4*t1^3*t2
        sage: e1 * e1 == e1
        True
        sage: e2 = Y.e(2); e2
        1/4 + 1/4*t2*t3^3 + 1/4*t2^2*t3^2 + 1/4*t2^3*t3
        sage: e2 * e2 == e2
        True

    REFERENCES:

    - [CL2013]_
    - [CPdA2014]_
    - [ERH2015]_
    - [JPdA15]_
    """
    def __init__(self, d, n, q, R):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(5, 3)
            sage: elts = Y.some_elements() + list(Y.algebra_generators())
            sage: TestSuite(Y).run(elements=elts)
        """
        self._n = n
        W = Permutations(n)
        import itertools
        C = itertools.product(*([range(d)]*n))
        indices = list(itertools.product(C, W))
        YokonumaHeckeAlgebra.__init__(self, d, W, q, R, indices)

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: algebras.YokonumaHecke(5, 2)
            Yokonuma-Hecke algebra of rank 5 and order 2 with q=q
             over Univariate Laurent Polynomial Ring in q over Rational Field
        """
        return "Yokonuma-Hecke algebra of rank {} and order {} with q={} over {}".format(
            self._d, self._n, self._q, self.base_ring())

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(5, 2)
            sage: latex(Y)
            \mathcal{Y}_{5,2}(q)
        """
        return "\\mathcal{Y}_{%s,%s}(%s)" % (self._d, self._n, self._q)

    def _repr_term(self, m):
        """
        Return a string representation of the basis element indexed by ``m``.

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(4, 3)
            sage: Y._repr_term( ((1, 0, 2), Permutation([3,2,1])) )
            't1*t3^2*g[2,1,2]'
        """
        def gen_str(e):
            return '' if e == 1 else '^%s' % e
        lhs = '*'.join('t%s' % (j+1) + gen_str(i) for j,i in enumerate(m[0]) if i > 0)
        redword = m[1].reduced_word()
        if not redword:
            if not lhs:
                return '1'
            return lhs
        rhs = 'g[{}]'.format(','.join(str(i) for i in redword))
        if not lhs:
            return rhs
        return lhs + '*' + rhs

    def _latex_term(self, m):
        r"""
        Return a latex representation for the basis element indexed by ``m``.

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(4, 3)
            sage: Y._latex_term( ((1, 0, 2), Permutation([3,2,1])) )
            't_{1} t_{3}^{2} g_{2} g_{1} g_{2}'
        """
        def gen_str(e):
            return '' if e == 1 else '^{%s}' % e
        lhs = ' '.join('t_{%s}' % (j+1) + gen_str(i) for j,i in enumerate(m[0]) if i > 0)
        redword = m[1].reduced_word()
        if not redword:
            if not lhs:
                return '1'
            return lhs
        return lhs + ' ' + ' '.join("g_{%d}" % i for i in redword)

    @cached_method
    def algebra_generators(self):
        """
        Return the algebra generators of ``self``.

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(5, 3)
            sage: dict(Y.algebra_generators())
            {'g1': g[1], 'g2': g[2], 't1': t1, 't2': t2, 't3': t3}
        """
        one = self._W.one()
        zero = [0] * self._n
        d = {}
        for i in range(self._n):
            r = list(zero) # Make a copy
            r[i] = 1
            d['t%s' % (i+1)] = self.monomial((tuple(r), one))
        G = self._W.group_generators()
        for i in range(1, self._n):
            d['g%s' % i] = self.monomial((tuple(zero), G[i]))
        return Family(sorted(d), lambda i: d[i])

    @cached_method
    def one_basis(self):
        """
        Return the index of the basis element of `1`.

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(5, 3)
            sage: Y.one_basis()
            ((0, 0, 0), [1, 2, 3])
        """
        one = self._W.one()
        zero = (0,) * self._n
        return (zero, one)

    @cached_method
    def e(self, i):
        """
        Return the element `e_i`.

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(4, 3)
            sage: Y.e(1)
            1/4 + 1/4*t1*t2^3 + 1/4*t1^2*t2^2 + 1/4*t1^3*t2
            sage: Y.e(2)
            1/4 + 1/4*t2*t3^3 + 1/4*t2^2*t3^2 + 1/4*t2^3*t3
        """
        if i < 1 or i >= self._n:
            raise ValueError("invalid index")
        c = ~self.base_ring()(self._d)
        zero = [0]*self._n
        one = self._W.one()
        d = {}
        for s in range(self._d):
            r = list(zero) # Make a copy
            r[i-1] = s
            if s != 0:
                r[i] = self._d - s
            d[(tuple(r), one)] = c
        return self._from_dict(d, remove_zeros=False)

    def t(self, i=None):
        """
        Return the generator(s) `t_i`.

        INPUT:

        - ``i`` -- (default: ``None``) the generator `t_i` or if ``None``,
          then the family of all generators `t_i`

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(8, 3)
            sage: Y.t(2)
            t2
            sage: list(Y.t())
            [t1, t2, t3]
        """
        G = self.algebra_generators()
        if i is None:
            I = tuple(range(1, self._n+1))
            d = {i: G['t%s' % i] for i in I}
            return Family(I, d.__getitem__)
        return G['t%s' % i]

    def product_on_basis(self, m1, m2):
        """
        Return the product of the basis elements indexed by ``m1`` and ``m2``.

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(4, 3)
            sage: m = ((1, 0, 2), Permutations(3)([2,1,3]))
            sage: 4 * Y.product_on_basis(m, m)
            -(q^-1-q)*t2^2*g[1] + 4*t1*t2 - (q^-1-q)*t1*t2*g[1]
             - (q^-1-q)*t1^2*g[1] - (q^-1-q)*t1^3*t2^3*g[1]

        Check that we apply the permutation correctly on `t_i`::

            sage: Y = algebras.YokonumaHecke(4, 3)
            sage: g1, g2, t1, t2, t3 = Y.algebra_generators()
            sage: g21 = g2 * g1
            sage: g21 * t1
            t3*g[2,1]
        """
        t1,g1 = m1
        t2,g2 = m2
        # Commute g1 and t2, then multiply t1 and t2
        # ig1 = g1
        t = [(t1[i] + t2[g1.index(i+1)]) % self._d for i in range(self._n)]
        one = self._W.one()
        if g1 == one:
            return self.monomial((tuple(t), g2))
        ret = self.monomial((tuple(t), g1))
        # We have to reverse the reduced word due to Sage's convention
        #   for permutation multiplication
        for i in g2.reduced_word():
            ret = self.linear_combination((self._product_by_basis_gen(m, i), c)
                                          for m,c in ret)
        return ret

    def _product_by_basis_gen(self, m, i):
        r"""
        Return the product `t g_w g_i`.

        If the quadratic relation is `g_i^2 = 1 + (q - q^{-1}) e_i g_i`,
        then we have

        .. MATH::

            g_w g_i = \begin{cases}
            g_{ws_i} & \text{if } \ell(ws_i) = \ell(w) + 1, \\
            g_{ws_i} + (q - q^{-1}) g_w e_i & \text{if }
            \ell(w s_i) = \ell(w) - 1.
            \end{cases}

        INPUT:

        - ``m`` -- a pair ``[t, w]``, where ``t`` encodes the monomial
          and ``w``  is an element of the permutation group
        - ``i`` -- an element of the index set

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(4, 3)
            sage: m = ((1, 0, 2), Permutations(3)([2,1,3]))
            sage: 4 * Y._product_by_basis_gen(m, 1)
            -(q^-1-q)*t2*t3^2*g[1] + 4*t1*t3^2 - (q^-1-q)*t1*t3^2*g[1]
             - (q^-1-q)*t1^2*t2^3*t3^2*g[1] - (q^-1-q)*t1^3*t2^2*t3^2*g[1]
        """
        t, w = m
        wi = w.apply_simple_reflection(i, side='right')
        if not w.has_descent(i, side='right'):
            return self.monomial((t, wi))

        R = self.base_ring()
        c = (self._q - ~self._q) * ~R(self._d)
        d = {(t, wi): R.one()}
        # We commute g_w and e_i and then multiply by t
        for s in range(self._d):
            r = list(t)
            r[w[i-1]-1] = (r[w[i-1]-1] + s) % self._d
            if s != 0:
                r[w[i]-1] = (r[w[i]-1] + self._d - s) % self._d
            d[(tuple(r), w)] = c
        return self._from_dict(d, remove_zeros=False)

    @cached_method
    def inverse_g(self, i):
        r"""
        Return the inverse of the generator `g_i`.

        From the quadratic relation, we have

        .. MATH::

            g_i^{-1} = g_i - (q - q^{-1}) e_i.

        INPUT:

        - ``i`` -- (default: ``None``) the inverse generator `g_i^{-1}` or
          if ``None``, then the family of all inverse generators `g_i^{-1}`

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(2, 4)
            sage: [2*Y.inverse_g(i) for i in range(1, 4)]
            [(q^-1-q) + 2*g[1] + (q^-1-q)*t1*t2,
             (q^-1-q) + 2*g[2] + (q^-1-q)*t2*t3,
             (q^-1-q) + 2*g[3] + (q^-1-q)*t3*t4]
            sage: all(Y.inverse_g(i) * Y.g(i) == Y.one() for i in range(1, 4))
            True
            sage: all(Y.g(i) * Y.inverse_g(i) == Y.one() for i in range(1, 4))
            True
        """
        if i is None:
            I = self._W.index_set()
            d = {i: self.inverse_g(i) for i in I}
            return Family(I, d.__getitem__)
        if i < 1 or i >= self._n:
            raise ValueError("invalid index")
        return self.g(i) + (~self._q - self._q) * self.e(i)

    class Element(CombinatorialFreeModule.Element):
        def __invert__(self):
            r"""
            Return the inverse if ``self`` is a basis element.

            EXAMPLES::

                sage: Y = algebras.YokonumaHecke(3, 3)
                sage: t = prod(Y.t()); t
                t1*t2*t3
                sage: t.inverse()   # indirect doctest
                t1^2*t2^2*t3^2
                sage: [3*~(t*g) for g in Y.g()]
                [(q^-1-q)*t2*t3^2 + (q^-1-q)*t1*t3^2
                   + (q^-1-q)*t1^2*t2^2*t3^2 + 3*t1^2*t2^2*t3^2*g[1],
                 (q^-1-q)*t1^2*t3 + (q^-1-q)*t1^2*t2
                   + (q^-1-q)*t1^2*t2^2*t3^2 + 3*t1^2*t2^2*t3^2*g[2]]
                sage: g = prod(Y.g())
                sage: ~g * g == Y.one()
                True
                sage: g * ~g == Y.one()
                True

                sage: tp = t * Y.t(2)
                sage: all(tp*g * ~(tp*g) == Y.one() for g in Y.g())
                True

            TESTS:

            Check that :issue:`26424` is fixed::

                sage: Y = algebras.YokonumaHecke(3, 3)
                sage: t = 3 * prod(Y.t())
                sage: ~t
                1/3*t1^2*t2^2*t3^2

                sage: ~Y.zero()
                Traceback (most recent call last):
                ...
                ZeroDivisionError
            """
            if not self:
                raise ZeroDivisionError
            if len(self) != 1:
                raise NotImplementedError("inverse only implemented for basis elements (monomials in the generators)" % self)
            H = self.parent()
            t, w = self.support_of_term()
            c = ~self.coefficients()[0]
            telt = H.monomial((tuple((H._d - e) % H._d for e in t), H._W.one()))
            return c * H.prod(H.inverse_g(i) for i in reversed(w.reduced_word())) * telt


class YokonumaHeckeAlgebraWeyl(YokonumaHeckeAlgebra):
    r"""
    The Yokonuma-Hecke algebra associated to a Cartan type.

    Let `R` be a commutative ring and `q` be a unit in `R`. Let
    `W` be the Weyl group acting on a root lattice `Q`. The
    *Yokonuma-Hecke algebra* `Y_{d,W}(q)` is the associative, unital
    `R`-algebra generated by `\{h_i, g_i \mid i \in I\}`, where `I` is
    the index set of simple roots of `Q`, and subject to the relations:

    - `g_i` and `g_j` satisfy the braid relations of the corresponding
      simple reflections `s_i` and `s_j` in `W`,
    - `h_i h_j = h_j h_i`,
    - `h_j g_i = g_i (s_i \cdot h_j)` with considering `h_j` as the simple
      root `\alpha_j \in Q`, and
    - `h_j^d = 1`,

    along with the quadratic relation

    .. MATH::

        g_i^2 = 1 + (q - 1) e_i (1 + g_i),
        \qquad\qquad
        e_i := \frac{1}{d} \sum_{s=0}^{d-1} h_i^s.

    In particular, we can identify the subalgebra generated by `\{h_i \mid
    i \in I\}` with `(\ZZ / d \ZZ) \otimes_{\ZZ} Q`. The Yokonuma-Hecke
    algebra, when `d = p^m - 1` for a prime `p` and some `m \geq 1`, can
    be identified with functions invariant under the left *and* right actions
    of the unipotent group `U` on `G(\GF{d})`, the semisimple Chevalley
    (or Lie) group associated with `W`. Moreover, all of the algebra
    generators are invertible. In particular, we have

    .. MATH::

        g_i^{-1} = g_i + (q^{-1} - 1) e_i (1 + g_i).

    For `d = 1`, the Yokonuma-Hecke algebra is equal to the
    :class:`Iwahori-Hecke <IwahoriHeckeAlgebra>` of `W`.

    INPUT:

    - ``d`` -- the maximum power of `t`
    - ``ct`` -- the Cartan type
    - ``q`` -- (optional) an invertible element in a commutative ring;
      the default is `q \in \QQ[q,q^{-1}]`
    - ``R`` -- (optional) a commutative ring containing ``q``; the
      default is the parent of `q`

    .. WARNING::

        For type `A_n`, this returns the Yokonuma-Hecke algebra associated
        to the Lie (or Chevalley) group `SL_n(\GF{d})`. For the Yokonuma-Hecke
        algebra corresponding to the (reductive) Lie group `GL_n(\GF{d})`, use
        :class:`YokonumaHeckeAlgebraGL`. Additionally, this uses a different
        quadratic relation.

    REFERENCES:

    - [Marin2018]_
    """
    def __init__(self, d, ct, q, R):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(2, ['F',4])
            sage: TestSuite(Y).run()
            sage: Y = algebras.YokonumaHecke(3, ['G',2])
            sage: elts = list(Y.gens()) + [Y.an_element()] + [sum(Y.gens())]
            sage: TestSuite(Y).run(elements=elts)  # long time
        """
        from sage.categories.sets_cat import cartesian_product
        from sage.rings.finite_rings.integer_mod_ring import IntegerModRing

        self._Q = ct.root_system().root_space(IntegerModRing(d))
        self._Qp = ct.root_system().root_lattice()
        W = self._Qp.weyl_group(prefix='s')
        indices = cartesian_product([self._Q, W])
        YokonumaHeckeAlgebra.__init__(self, d, W, q, R, indices)

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: algebras.YokonumaHecke(5, ['E',6])
            Yokonuma-Hecke algebra of rank 5 for ['E', 6] with q=q
             over Univariate Laurent Polynomial Ring in q over Rational Field
        """
        return "Yokonuma-Hecke algebra of rank {} for {} with q={} over {}".format(
            self._d, self._cartan_type, self._q, self.base_ring())

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(5, ['E',6])
            sage: latex(Y)
            \mathcal{Y}_{5,E_6}(q)
        """
        from sage.misc.latex import latex
        return "\\mathcal{Y}_{%s,%s}(%s)" % (self._d, latex(self._cartan_type), self._q)

    def _repr_term(self, m):
        """
        Return a string representation of the basis element indexed by ``m``.

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(4, ['E',6])
            sage: al = Y._Q.simple_root(1) + 3*Y._Q.simple_root(5)
            sage: Y._repr_term((al, prod(Y._W.gens())))
            'h1*h5^3*g[1,3,2,4,5,6]'
        """
        def gen_str(e):
            return '' if e == 1 else '^%s' % e

        I = self._cartan_type.index_set()
        lhs = '*'.join('h%s' % j + gen_str(m[0][j]) for j in I if m[0][j])
        redword = m[1].reduced_word()
        if not redword:
            if not lhs:
                return '1'
            return lhs
        rhs = 'g[{}]'.format(','.join(str(i) for i in redword))
        if not lhs:
            return rhs
        return lhs + '*' + rhs

    def _latex_term(self, m):
        r"""
        Return a latex representation for the basis element indexed by ``m``.

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(4, ['E',6])
            sage: al = Y._Q.simple_root(1) + 3*Y._Q.simple_root(5)
            sage: Y._latex_term((al, prod(Y._W.gens())))
            'h_{1} h_{5}^{3} g_{1} g_{3} g_{2} g_{4} g_{5} g_{6}'
        """
        def gen_str(e):
            return '' if e == 1 else '^{%s}' % e

        I = self._cartan_type.index_set()
        lhs = ' '.join('h_{%s}' % j + gen_str(m[0][j]) for j in I if m[0][j])
        redword = m[1].reduced_word()
        if not redword:
            if not lhs:
                return '1'
            return lhs
        return lhs + ' ' + ' '.join("g_{%d}" % i for i in redword)

    @cached_method
    def algebra_generators(self):
        """
        Return the algebra generators of ``self``.

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(5, ['G',2])
            sage: dict(Y.algebra_generators())
            {'g1': g[1], 'g2': g[2], 'h1': h1, 'h2': h2}
        """
        one = self._W.one()
        zero = self._Q.zero()
        d = {}
        for i, al in self._Q.simple_roots().items():
            d['h%s' % i] = self.monomial((al, one))
        for i, g in self._W.simple_reflections().items():
            d['g%s' % i] = self.monomial((zero, g))
        return Family(sorted(d), d.__getitem__)

    @cached_method
    def one_basis(self):
        """
        Return the index of the basis element of `1`.

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(5, ['D',6])
            sage: Y.one_basis()
            (0, 1)
        """
        return (self._Q.zero(), self._W.one())

    @cached_method
    def e(self, i=None):
        r"""
        Return the element(s) `e_i`.

        INPUT:

        - ``i`` -- (default: ``None``) the element `e_i` or if ``None``,
          then the family of all idempotents `e_i`

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(4, ['B',3])
            sage: Y.e(1)
            1/4 + 1/4*h1 + 1/4*h1^2 + 1/4*h1^3
            sage: Y.e(2)
            1/4 + 1/4*h2 + 1/4*h2^2 + 1/4*h2^3

        We verify that they are idempotents::

            sage: all(Y.e(i)^2 == Y.e(i) for i in Y.index_set())
            True

        Another example::

            sage: Y = algebras.YokonumaHecke(3, ['G',2])
            sage: e = Y.e()
            sage: all(e[i]^2 == e[i] for i in Y.index_set())
            True
        """
        if i is None:
            I = self._W.index_set()
            d = {i: self.e(i) for i in I}
            return Family(I, d.__getitem__)
        if i not in self._W.index_set():
            raise ValueError("invalid index")
        c = ~self.base_ring()(self._d)
        al = self._Q.simple_root(i)
        one = self._W.one()
        d = {(k*al, one): c for k in self._Q.base_ring()}
        return self._from_dict(d, remove_zeros=False)

    def h(self, i=None):
        r"""
        Return the generator(s) `h_i`.

        INPUT:

        - ``i`` -- (default: ``None``) the generator `h_i` or if ``None``,
          then the family of all generators `h_i`

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(8, ['B',3])
            sage: Y.h(2)
            h2
            sage: Y.h()
            Finite family {1: h1, 2: h2, 3: h3}
        """
        G = self.algebra_generators()
        if i is None:
            I = self._W.index_set()
            d = {i: G['h%s' % i] for i in I}
            return Family(I, d.__getitem__)
        return G['h%s' % i]

    @cached_method
    def inverse_g(self, i=None):
        r"""
        Return the inverse of the generator(s) `g_i`.

        From the quadratic relation, we have

        .. MATH::

            g_i^{-1} = g_i + (q^{-1} - 1) e_i (1 + g_i).

        INPUT:

        - ``i`` -- (default: ``None``) the inverse generator `g_i^{-1}` or
          if ``None``, then the family of all inverse generators `g_i^{-1}`

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(2, ['B',3])
            sage: [2*Y.inverse_g(i) for i in Y.index_set()]
            [(q^-1+1)*g[1] + (q^-1-1) + (q^-1-1)*h1*g[1] + (q^-1-1)*h1,
             (q^-1-1) + (q^-1+1)*g[2] + (q^-1-1)*h2 + (q^-1-1)*h2*g[2],
             (q^-1-1) + (q^-1+1)*g[3] + (q^-1-1)*h3 + (q^-1-1)*h3*g[3]]
            sage: all(Y.inverse_g(i) * Y.g(i) == Y.one() for i in range(1, 4))
            True
            sage: all(Y.g(i) * Y.inverse_g(i) == Y.one() for i in range(1, 4))
            True

            sage: Y = algebras.YokonumaHecke(3, ['G',2])
            sage: ginv = Y.inverse_g()
            sage: all(Y.g(i) * ginv[i] == Y.one() for i in Y.index_set())
            True
            sage: all(ginv[i] * Y.g(i) == Y.one() for i in Y.index_set())
            True
        """
        if i is None:
            I = self._W.index_set()
            d = {i: self.inverse_g(i) for i in I}
            return Family(I, d.__getitem__)
        if i not in self._W.index_set():
            raise ValueError("invalid index")
        return self.g(i) + (~self._q - 1) * self.e(i) * (self.one() + self.g(i))

    def product_on_basis(self, m1, m2):
        r"""
        Return the product of the basis elements indexed by ``m1`` and ``m2``.

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(4, ['C',3])
            sage: al = Y._Q.simple_root(1) + 2*Y._Q.simple_root(3)
            sage: w = Y._W.from_reduced_word([3,2,1,2]); w.length()
            4
            sage: Y.product_on_basis((Y._Q.zero(), w), (al, Y._W.one()))
            h2^3*h3*g[3,1,2,1]
            sage: Y.product_on_basis((al, w), (al, Y._W.one()))
            h1*h2^3*h3^3*g[3,1,2,1]
            sage: Y.product_on_basis((al, Y._W.one()), (al, w))
            h1^2*g[3,1,2,1]
            sage: 4 * Y.product_on_basis((al, w), (al, w))
            -(1-q)*h1*g[3,1,2,3,2,1] - (1-q)*h1*g[3,1,2,3,1,2,1]
             - (1-q)*h1*h2*h3*g[3,1,2,3,2,1] - (1-q)*h1*h2*h3*g[3,1,2,3,1,2,1]
             - (1-q)*h1*h2^2*h3^2*g[3,1,2,3,2,1] - (1-q)*h1*h2^2*h3^2*g[3,1,2,3,1,2,1]
             + (3+q)*h1*h2^3*h3^3*g[3,1,2,3,2,1] - (1-q)*h1*h2^3*h3^3*g[3,1,2,3,1,2,1]

        Check that we apply the permutation correctly on `h_i`::

            sage: Y = algebras.YokonumaHecke(4, ['B',3])
            sage: g1, g2, g3, h1, h2, h3 = Y.algebra_generators()
            sage: (g2 * g1) * h1
            h1^3*h2^3*g[2,1]
            sage: g2 * (g1 * h1)
            h1^3*h2^3*g[2,1]
        """
        h1, g1 = m1
        h2, g2 = m2
        # Commute g1 and t2, then multiply h1 and h2
        # ig1 = g1
        h = h1 + h2.weyl_action(g1)
        one = self._W.one()
        if g1 == one:
            return self.monomial((h, g2))
        ret = self.monomial((h, g1))
        for i in g2.reduced_word():
            ret = self.linear_combination((self._product_by_basis_gen(m, i), c)
                                          for m, c in ret)
        return ret

    def _product_by_basis_gen(self, m, i):
        r"""
        Return the product `t g_w g_i`.

        If the quadratic relation is `g_i^2 = 1 + (q - 1) (1 + g_i) e_i`,
        then we have

        .. MATH::

            g_w g_i = \begin{cases}
            g_{ws_i} & \text{if } \ell(ws_i) = \ell(w) + 1, \\
            g_{ws_i} + (q-1) (g_{ws_i} + g_w) e_i & \text{if }
              \ell(w s_i) = \ell(w) - 1.
            \end{cases}

        INPUT:

        - ``m`` -- a pair ``[h, w]``, where ``h`` encodes the monomial
          and ``w``  is an element of the Weyl group
        - ``i`` -- an element of the index set

        EXAMPLES::

            sage: Y = algebras.YokonumaHecke(4, ['D',4])
            sage: m = ((1, 0, 2), Permutations(3)([2,1,3]))
            sage: 4 * Y._product_by_basis_gen(m, 1)  # not tested
            -(q^-1-q)*t2*t3^2*g[1] + 4*t1*t3^2 - (q^-1-q)*t1*t3^2*g[1]
             - (q^-1-q)*t1^2*t2^3*t3^2*g[1] - (q^-1-q)*t1^3*t2^2*t3^2*g[1]
        """
        h, w = m
        wi = w.apply_simple_reflection(i, side='right')
        if not w.has_descent(i, side='right'):
            return self.monomial((h, wi))

        q = self._q
        mon = self.monomial((h, wi))
        # TODO: Optimize this by computing an explicit expression
        #   for the commutation of w with ei.
        one = self.base_ring().one()
        binomial = self.element_class(self, {(h,wi): one, (h,w): one})
        return mon + (q-1) * binomial * self.e(i)

    class Element(CombinatorialFreeModule.Element):
        def __invert__(self):
            r"""
            Return the inverse if ``self`` is a basis element.

            EXAMPLES::

                sage: Y = algebras.YokonumaHecke(3, ['B',3])
                sage: all(g * ~g == Y.one() for g in Y.g())
                True
                sage: h = prod(Y.h()) * Y.h(2)
                sage: all(h*g * ~(h*g) == Y.one() for g in Y.g())
                True
            """
            if not self:
                raise ZeroDivisionError
            if len(self) != 1:
                raise NotImplementedError("inverse only implemented for basis elements (monomials in the generators)" % self)
            H = self.parent()
            t, w = self.support_of_term()
            c = ~self.coefficients()[0]
            telt = H.monomial((-t, H._W.one()))
            return c * H.prod(H.inverse_g(i) for i in reversed(w.reduced_word())) * telt



# sage_setup: distribution = sagemath-combinat
r"""
Infinite word

AUTHORS:

- Sebastien Labbe
- Franco Saliola

EXAMPLES:

============================
Creation of an infinite word
============================

Periodic infinite words::

    sage: v = Word([0, 4, 8, 8, 3])
    sage: vv = v^Infinity
    sage: vv
    word: 0488304883048830488304883048830488304883...

Infinite words from a function `f:\mathbb{N}\rightarrow A`
over an alphabet `A`::

    sage: Word(lambda n: n%3)
    word: 0120120120120120120120120120120120120120...

::

    sage: def t(n):
    ....:     return add(Integer(n).digits(base=2)) % 2
    sage: Word(t, alphabet = [0, 1])
    word: 0110100110010110100101100110100110010110...

or as a one-liner::

    sage: Word(lambda n : add(Integer(n).digits(base=2)) % 2, alphabet = [0, 1])
    word: 0110100110010110100101100110100110010110...

Infinite words from iterators::

    sage: from itertools import count,repeat
    sage: Word( repeat(4) )
    word: 4444444444444444444444444444444444444444...
    sage: Word( count() )
    word: 0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,...

Infinite words from morphism

For example, let `A=\{a,b\}` and `\mu : A^* \rightarrow A^*`
be the morphism defined by `a\mapsto ab, b\mapsto ba`::

    sage: mu = WordMorphism('a->ab,b->ba'); mu
    WordMorphism: a->ab, b->ba
    sage: mu.fixed_point('a')
    word: abbabaabbaababbabaababbaabbabaabbaababba...

Infinite words in a specific combinatorial class::

    sage: W = InfiniteWords("ab"); W
    Infinite words over {'a', 'b'}
    sage: f = lambda n : 'a' if n % 2 == 1 else 'b'
    sage: W(f)
    word: babababababababababababababababababababa...
"""
# ****************************************************************************
#       Copyright (C) 2008 Sebastien Labbe <slabqc@gmail.com>,
#                          Franco Saliola <saliola@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from sage.combinat.words.abstract_word import Word_class
from sage.combinat.words.word_options import word_options
from sage.rings.infinity import Infinity


class InfiniteWord_class(Word_class):
    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        TESTS::

            sage: Word(iter([1,2,3]), length='unknown')._repr_()
            'word: 123'
            sage: Word(range(100), length='unknown')._repr_()
            'word: 0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,...'
            sage: Word(lambda x:x%3)._repr_()
            'word: 0120120120120120120120120120120120120120...'
        """
        global word_options
        if word_options['old_repr']:
            return "Infinite word over %s" % str(self.parent().alphabet())[17:]
        return word_options['identifier'] + self.string_rep()

    def length(self):
        r"""
        Return the length of ``self``.

        EXAMPLES::

            sage: f = lambda n : n % 6
            sage: w = Word(f); w
            word: 0123450123450123450123450123450123450123...
            sage: w.length()
            +Infinity
        """
        return Infinity



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.graphs sage.modules
r"""
Affinization crystals
"""

#*****************************************************************************
#       Copyright (C) 2015 Travis Scrimshaw <tscrim at ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#****************************************************************************

from sage.structure.parent import Parent
from sage.structure.unique_representation import UniqueRepresentation
from sage.structure.element import Element
from sage.structure.richcmp import richcmp
from sage.categories.regular_crystals import RegularCrystals
from sage.categories.infinite_enumerated_sets import InfiniteEnumeratedSets
from sage.rings.infinity import Infinity


class AffinizationOfCrystal(UniqueRepresentation, Parent):
    r"""
    An affinization of a crystal.

    Let `\mathfrak{g}` be a Kac-Moody algebra of affine type. The
    affinization of a finite `U_q^{\prime}(\mathfrak{g})`-crystal `B`
    is the (infinite) `U_q(\mathfrak{g})`-crystal with underlying set:

    .. MATH::

        B^{\mathrm{aff}} = \{ b(m) \mid b \in B, m \in \ZZ \}

    and crystal structure determined by:

    .. MATH::

        \begin{aligned}
            e_i(b(m)) & =
            \begin{cases}
              (e_0 b)(m+1) & i = 0, \\
              (e_i b)(m)   & i \neq 0,
            \end{cases} \\
            f_i(b(m)) &=
            \begin{cases}
              (f_0 b)(m-1) & i = 0, \\
              (f_i b)(m)   & i \neq 0,
            \end{cases} \\
            \mathrm{wt}(b(m)) &= \mathrm{wt}(b) + m \delta.
        \end{aligned}

    EXAMPLES:

    We first construct a Kirillov-Reshetikhin crystal and then take it's
    corresponding affinization::

        sage: K = crystals.KirillovReshetikhin(['A',2,1], 2, 2)
        sage: A = K.affinization()

    Next we construct an affinization crystal from a tensor product of KR
    crystals::

        sage: KT = crystals.TensorProductOfKirillovReshetikhinTableaux(['C',2,1], [[1,2],[2,1]])
        sage: A = crystals.AffinizationOf(KT)

    REFERENCES:

    - [HK2002]_ Chapter 10
    """

    def __init__(self, B):
        """
        Initialize ``self``.

        EXAMPLES:

        We skip the Stembridge axioms test since this is an abstract crystal::

            sage: A = crystals.KirillovReshetikhin(['A',2,1], 2, 2).affinization()
            sage: TestSuite(A).run(skip='_test_stembridge_local_axioms') # long time
        """
        if not B.cartan_type().is_affine():
            raise ValueError("must be an affine crystal")
        if B.cardinality() == Infinity:
            raise ValueError("must be finite crystal")
        self._B = B
        self._cartan_type = B.cartan_type()
        Parent.__init__(self, category=(RegularCrystals(), InfiniteEnumeratedSets()))
        self.module_generators = tuple([self.element_class(self, b, 0)
                                        for b in B.module_generators])

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: crystals.KirillovReshetikhin(['A',2,1], 1, 1).affinization()
            Affinization of Kirillov-Reshetikhin crystal of type ['A', 2, 1] with (r,s)=(1,1)
        """
        return "Affinization of {}".format(self._B)

    class Element(Element):
        """
        An element in an affinization crystal.
        """

        def __init__(self, parent, b, m):
            """
            Initialize ``self``.

            EXAMPLES::

                sage: A = crystals.KirillovReshetikhin(['A',2,1], 2, 2).affinization()
                sage: mg = A.module_generators[0]
                sage: TestSuite(mg).run()
            """
            self._b = b
            self._m = m
            Element.__init__(self, parent)

        def _repr_(self):
            """
            Return a string representation of ``self``.

            EXAMPLES::

                sage: A = crystals.KirillovReshetikhin(['A',2,1], 2, 2).affinization()
                sage: A.module_generators[0]
                [[1, 1], [2, 2]](0)
                sage: KT = crystals.TensorProductOfKirillovReshetikhinTableaux(['C',2,1], [[1,2],[2,1]])
                sage: A = crystals.AffinizationOf(KT)
                sage: A.module_generators[0]
                [[1, 1]] (X) [[1], [2]](0)
            """
            return "{!r}({})".format(self._b, self._m)

        def _latex_(self):
            r"""
            Return a LaTeX representation of ``self``.

            EXAMPLES::

                sage: A = crystals.KirillovReshetikhin(['A',2,1], 2, 2).affinization()
                sage: latex(A.module_generators[0])
                {\def\lr#1{\multicolumn{1}{|@{\hspace{.6ex}}c@{\hspace{.6ex}}|}{\raisebox{-.3ex}{$#1$}}}
                \raisebox{-.6ex}{$\begin{array}[b]{*{2}c}\cline{1-2}
                \lr{1}&\lr{1}\\\cline{1-2}
                \lr{2}&\lr{2}\\\cline{1-2}
                \end{array}$}
                } (0)
            """
            from sage.misc.latex import latex
            return latex(self._b) + "({})".format(self._m)

        def __hash__(self):
            r"""
            TESTS::

                sage: A = crystals.KirillovReshetikhin(['A',2,1], 2, 2).affinization()
                sage: mg = A.module_generators[0]
                sage: hash(mg) == hash(mg._b) ^^ hash(mg._m)
                True
            """
            return hash(self._b) ^ hash(self._m)

        def _richcmp_(self, other, op):
            """
            Comparison.

            TESTS::

                sage: A = crystals.KirillovReshetikhin(['A',2,1], 2, 2).affinization()
                sage: mg = A.module_generators[0]
                sage: mg == mg
                True
                sage: mg == mg.f(2).e(2)
                True
                sage: KT = crystals.TensorProductOfKirillovReshetikhinTableaux(['C',2,1], [[1,2],[2,1]])
                sage: A = crystals.AffinizationOf(KT)
                sage: A(KT.module_generators[3], 1).f(0) == A.module_generators[0]
                True

                sage: A = crystals.KirillovReshetikhin(['A',2,1], 2, 2).affinization()
                sage: mg = A.module_generators[0]
                sage: mg != mg.f(2)
                True
                sage: mg != mg.f(2).e(2)
                False


                sage: A = crystals.KirillovReshetikhin(['A',2,1], 2, 2).affinization()
                sage: S = A.subcrystal(max_depth=2)
                sage: sorted(S)
                [[[1, 1], [2, 2]](0),
                 [[1, 1], [2, 3]](0),
                 [[1, 2], [2, 3]](0),
                 [[1, 1], [3, 3]](0),
                 [[1, 1], [2, 3]](1),
                 [[1, 2], [2, 3]](1),
                 [[1, 2], [3, 3]](1),
                 [[2, 2], [3, 3]](2)]
            """
            return richcmp((self._m, self._b), (other._m, other._b), op)

        def e(self, i):
            """
            Return the action of `e_i` on ``self``.

            INPUT:

            - ``i`` -- an element of the index set

            EXAMPLES::

                sage: A = crystals.KirillovReshetikhin(['A',2,1], 2,2).affinization()
                sage: mg = A.module_generators[0]
                sage: mg.e(0)
                [[1, 2], [2, 3]](1)
                sage: mg.e(1)
                sage: mg.e(0).e(1)
                [[1, 1], [2, 3]](1)
            """
            bp = self._b.e(i)
            if bp is None:
                return None
            if i == 0:
                return self.__class__(self.parent(), bp, self._m+1)
            return self.__class__(self.parent(), bp, self._m)

        def f(self, i):
            """
            Return the action of `f_i` on ``self``.

            INPUT:

            - ``i`` -- an element of the index set

            EXAMPLES::

                sage: A = crystals.KirillovReshetikhin(['A',2,1], 2,2).affinization()
                sage: mg = A.module_generators[0]
                sage: mg.f(2)
                [[1, 1], [2, 3]](0)
                sage: mg.f(2).f(2).f(0)
                sage: mg.f_string([2,1,1])
                sage: mg.f_string([2,1])
                [[1, 2], [2, 3]](0)
                sage: mg.f_string([2,1,0])
                [[1, 1], [2, 2]](-1)
            """
            bp = self._b.f(i)
            if bp is None:
                return None
            if i == 0:
                return self.__class__(self.parent(), bp, self._m-1)
            return self.__class__(self.parent(), bp, self._m)

        def epsilon(self, i):
            r"""
            Return `\varepsilon_i` of ``self``.

            INPUT:

            - ``i`` -- an element of the index set

            EXAMPLES::

                sage: A = crystals.KirillovReshetikhin(['A',2,1], 2,2).affinization()
                sage: mg = A.module_generators[0]
                sage: mg.epsilon(0)
                2
                sage: mg.epsilon(1)
                0
            """
            return self._b.epsilon(i)

        def phi(self, i):
            r"""
            Return `\varphi_i` of ``self``.

            INPUT:

            - ``i`` -- an element of the index set

            EXAMPLES::

                sage: A = crystals.KirillovReshetikhin(['A',2,1], 2,2).affinization()
                sage: mg = A.module_generators[0]
                sage: mg.phi(0)
                0
                sage: mg.phi(2)
                2
            """
            return self._b.phi(i)

        def weight(self):
            r"""
            Return the weight of ``self``.

            The weight `\mathrm{wt}` of an element is:

            .. MATH::

                \mathrm{wt}\bigl( b(m) \bigr) = \mathrm{wt}(b) + m \delta,

            where `\delta` is the null root.

            EXAMPLES::

                sage: A = crystals.KirillovReshetikhin(['A',2,1], 2,2).affinization()
                sage: mg = A.module_generators[0]
                sage: mg.weight()
                -2*Lambda[0] + 2*Lambda[2]
                sage: mg.e(0).weight()
                -Lambda[1] + Lambda[2] + delta
                sage: mg.e(0).e(0).weight()
                2*Lambda[0] - 2*Lambda[1] + 2*delta
            """
            WLR = self.parent().weight_lattice_realization()
            La = WLR.fundamental_weights()
            return WLR.sum(c*La[i] for i,c in self._b.weight()) + self._m * WLR.null_root()
